# Contextual refinement of the Michael Scott queue

This file contains the Coq development accompanying the paper "Contextual
Refinement of the Michael-Scott Queue (Proof Pearl)".

Note that this archive contains both our refinement proofs and forks of Iris and
of ReLoC extended with support for our persistent points-to predicate.

Our refinement proofs are in the directory `theories` and the forks are in
`iris` and `reloc`.

Note, in the Coq development the persistent points-to predicate is denoted `l ↦□
v`.

## Build instructions

The development has been done and tested using Coq v. 8.12.0.

First, follow the instructions in the section "Working _on_ Iris" in
`iris/readme.md`. This installs the dependencies for Iris.

Then run
```
make
```
in the root (where this README is located).

This will compile the Coq files in the `iris`, `reloc`, and `theories`
directory. After this it should also be possible to load the Coq files in an
editor capable of using the `_CoqProject` file (for instance Emacs or VSCode).

## Overview of the files

In the following sections we describe the files in `theories` and then some of
the changes we made to Iris and ReLoC.

### Files for the MS-queue

These are all in the `theories` subdirectory.

The HeapLang implementations of the data structures:

- `CG_queue`: Contains the HeapLang implementation of the coarse-grained queue.
- `MS_queue`: Contains the HeapLang implementation of the original unchanged
  MS-queue.
- `MS_queue_no_snapshot`: Contains the HeapLang implementation of the MS-queue
  with the consistent snapshots removed.
- `MS_queue_lagging`: Contains the HeapLang implementation of the lagging tail
  MS-queue.

The file `extra_rules.v` contains a few extra rules for the refinement judgment.

The refinement proofs:

- `MS_refines_CG.v`: The MS-queue refines the coarse-grained queue.
- `MS_no_snapshot_refines_CG.v`: The MS-queue with no snapshots refines the coarse-grained queue.
- `MS_lagging_refines_CG.v`: The lagging tail MS-queue refines the coarse-grained queue.
- `CG_refines_MS.v`: The coarse-grained queue refines the MS-queue.
- `CG_refines_MS_no_snapshot.v`: The coarse-grained queue refines the MS-queue.

Finally `equivalence.v` contains the equivalence proofs based on the above.

### Iris with persistent points-to predicate

The directory `iris` contains our fork of Iris with support for the persistent
points-to predicate.

A full diff of all the changes we made to Iris can be found in: `iris.diff`.

We have made the following additions and changes:

- We have implemented the two resource algebras described in the paper.
  - The discardable fractions resource algebra is in `iris/theories/algebra/dfrac.v`.
  - The authorative resource algebra with projections is in `iris/theories/algebra/pauth.v`.
- To implement the persistent points-to predicate we have, among other things,
  modified:
  - `theories/base_logic/lib/gen_heap.v`: The encoding of the heap.
  - `theories/heap_lang/primitive_laws.v`: Primitive rules for HeapLang.
  - `theories/heap_lang/proofmode.v`: Tactics for the proof mode.

### ReLoC with persistent points-to predicate

The directory `reloc` contains our fork of ReLoC with support for the persistent
points-to predicate. We have made the following additions and changes:

A full diff of all the changes we made to ReLoC can be found in: `reloc.diff`.

- `theories/logic/proofmode/tactics.v`: ReLoC tactics.
- `theories/logic/rules.v`: Adde a load rule for the persistent points-to
  predicate.
- `theories/logic/spec_ra.v`: Adapted the ReLoC heap encoding to support the
  persistent points-to predicate.
