From iris.proofmode Require Import tactics.
From iris.algebra Require Export dfrac agree excl gmap.
From iris.algebra Require Import proofmode_classes.
From iris.base_logic Require Import base_logic.
Set Default Proof Using "Type".

Record dauth `{Countable K} (A : cmraT) :=
  DAuth {
    dauth_auth_proj : option (excl (gmap K A)) ;
    dauth_frag_proj : gmap K (prodR dfracR A)
  }.
Add Printing Constructor dauth.
Check dauth.
Arguments dauth _ {_ _} _.
Arguments DAuth {_} {_} {_} {_} _ _.
Arguments dauth_auth_proj {_ _ _ _} _.
Arguments dauth_frag_proj {_ _ _ _} _.
Instance: Params (@DAuth) 1 := {}.
Instance: Params (@dauth_auth_proj) 1 := {}.
Instance: Params (@dauth_frag_proj) 1 := {}.

Definition dauth_frag `{Countable K} {A : cmraT} (a : gmap K (prodR dfracR A)) :=
  DAuth None a.
Definition dauth_auth `{Countable K} {A : cmraT} (a : gmap K A) :=
  DAuth (Some (Excl a)) ε.

Typeclasses Opaque dauth_auth dauth_frag.

Instance: Params (@dauth_frag) 1 := {}.
Instance: Params (@dauth_auth) 1 := {}.

Notation "◯ a" := (dauth_frag a) (at level 20).
Notation "● a" := (dauth_auth a) (at level 20).

(* Ofe *)
Section ofe.
Context `{Countable K} {A : cmraT}.

Check dauth.
Instance dauth_equiv : Equiv (dauth K A) := λ x y,
  dauth_auth_proj x ≡ dauth_auth_proj y ∧ dauth_frag_proj x ≡ dauth_frag_proj y.
Instance dauth_dist : Dist (dauth K A) := λ n x y,
  dauth_auth_proj x ≡{n}≡ dauth_auth_proj y ∧
  dauth_frag_proj x ≡{n}≡ dauth_frag_proj y.

Global Instance DAuth_ne : NonExpansive2 (@DAuth K _ _ A).
Proof. by split. Qed.
Global Instance DAuth_proper : Proper ((≡) ==> (≡) ==> (≡)) (@DAuth K _ _ A).
Proof. by split. Qed.
Global Instance dauth_auth_proj_ne: NonExpansive (@dauth_auth_proj K _ _ A).
Proof. by destruct 1. Qed.
Global Instance dauth_auth_proj_proper : Proper ((≡) ==> (≡)) (@dauth_auth_proj K _ _ A).
Proof. by destruct 1. Qed.
Global Instance dauth_frag_proj_ne : NonExpansive (@dauth_frag_proj K _ _ A).
Proof. by destruct 1. Qed.
Global Instance dauth_frag_proj_proper : Proper ((≡) ==> (≡)) (@dauth_frag_proj K _ _ A).
Proof. by destruct 1. Qed.

Definition dauth_ofe_mixin : OfeMixin (dauth K A).
Proof. by apply (iso_ofe_mixin (λ x, (dauth_auth_proj x, dauth_frag_proj x))). Qed.
Canonical Structure dauthO := OfeT (dauth K A) dauth_ofe_mixin.

Global Instance DAuth_discrete a b :
  Discrete a → Discrete b → Discrete (DAuth a b).
Proof. by intros ?? [??] [??]; split; apply: discrete. Qed.
Global Instance dauth_ofe_discrete : OfeDiscrete A → OfeDiscrete dauthO.
Proof. intros ? [??]; apply _. Qed.
End ofe.

Arguments dauthO : clear implicits.

(* Camera *)
Section cmra.
Context `{Countable K} {A : ucmraT}.
(* Implicit Types a b : A. *)
Implicit Types a : gmap K A.
Implicit Types b : gmap K (dfrac * A).
Implicit Types x y : dauth K A.

Global Instance dauth_frag_ne: NonExpansive (@dauth_frag K _ _ A).
Proof. done. Qed.
Global Instance dauth_frag_proper : Proper ((≡) ==> (≡)) (@dauth_frag K _ _ A).
Proof. done. Qed.
Global Instance dauth_auth_ne : NonExpansive (@dauth_auth K _ _ A).
Proof. solve_proper. Qed.
Global Instance dauth_auth_proper : Proper ((≡) ==> (≡)) (@dauth_auth K _ _ A).
Proof. solve_proper. Qed.
Global Instance dauth_auth_discrete a :
  Discrete a → Discrete (ε : A) → Discrete (● a).
Proof. intros. apply DAuth_discrete; apply _. Qed.
Global Instance dauth_frag_discrete b : Discrete b → Discrete (◯ b).
Proof. intros. apply DAuth_discrete; apply _. Qed.

Instance dauth_valid : Valid (dauth K A) := λ x,
  match x with
  | DAuth (Some (Excl a)) b => ✓ a ∧ (snd <$> b) ≼ a
      (* ✓ a ∧ (∀ n, ∃ a, ag ≡{n}≡ to_agree a ∧ dauth_frag_proj x ≼{n} a ∧ ✓{n} a) *)
  | DAuth None b => ✓ b
  | _ => False
  end.
Global Arguments dauth_valid !_ /.
Instance dauth_validN : ValidN (dauth K A) := λ n x,
  match x with
  | DAuth (Some (Excl a)) b => ✓ a ∧ (snd <$> b) ≼ a
  | DAuth None b => ✓ b
  | _ => False
  end.
Global Arguments dauth_validN _ !_ /.
Instance dauth_pcore : PCore (dauth K A) := λ 'x,
  Some (DAuth None (core (dauth_frag_proj x))).
Instance dauth_op : Op (dauth K A) := λ x y,
  DAuth (dauth_auth_proj x ⋅ dauth_auth_proj y) (dauth_frag_proj x ⋅ dauth_frag_proj y).

Lemma dauth_included (x y : dauth K A) :
  x ≼ y ↔
  dauth_auth_proj x ≼ dauth_auth_proj y ∧ dauth_frag_proj x ≼ dauth_frag_proj y.
Proof.
  split; [intros [[z1 z2] Hz]; split; [exists z1|exists z2]; apply Hz|].
  intros [[z1 Hz1] [z2 Hz2]]; exists (DAuth z1 z2); split; auto.
Qed.

(* Lemma dauth_auth_proj_validN n x : ✓{n} x → ✓{n} dauth_auth_proj x.
Proof. destruct x as [[[]|]]; [by intros (? & ? & -> & ?)|done]. Qed. *)
(* Lemma dauth_frag_proj_validN n x : ✓{n} x → ✓{n} dauth_frag_proj x.
Proof.
  rewrite dauth_validN_eq.
  destruct x as [[[]|]]; [|done]. naive_solver eauto using cmra_validN_includedN.
Qed. *)
(* Lemma dauth_auth_proj_valid x : ✓ x → ✓ dauth_auth_proj x.
Proof.
  destruct x as [[[]|]]; [intros [? H]|done]. split; [done|].
  intros n. by destruct (H n) as [? [-> [? ?]]].
Qed. *)
(* Lemma dauth_frag_proj_valid x : ✓ x → ✓ dauth_frag_proj x.
Proof.
  destruct x as [[[]|]]; [intros [? H]|done]. apply cmra_valid_validN.
  intros n. destruct (H n) as [? [? [? ?]]].
  naive_solver eauto using cmra_validN_includedN.
Qed. *)

(* Lemma dauth_frag_validN n a : ✓{n} (◯ a) ↔ ✓{n} a.
Proof. done. Qed. *)
(* Lemma dauth_auth_frac_validN n q a :
  ✓{n} (●{q} a) ↔ ✓{n} q ∧ ✓{n} a.
Proof.
  rewrite dauth_validN_eq /=. apply and_iff_compat_l. split.
  - by intros [?[->%(inj to_agree) []]].
  - naive_solver eauto using ucmra_unit_leastN.
Qed. *)
(* Lemma dauth_auth_validN n a : ✓{n} a ↔ ✓{n} (● a).
Proof.
  rewrite dauth_auth_frac_validN -cmra_discrete_valid_iff frac_valid'. naive_solver.
Qed. *)
(* Lemma dauth_both_frac_validN n q a b :
  ✓{n} (●{q} a ⋅ ◯ b) ↔ ✓{n} q ∧ b ≼{n} a ∧ ✓{n} a.
Proof.
  rewrite dauth_validN_eq /=. apply and_iff_compat_l.
  setoid_rewrite (left_id _ _ b). split.
  - by intros [?[->%(inj to_agree)]].
  - naive_solver.
Qed. *)
(* Lemma dauth_both_validN n a b : ✓{n} (● a ⋅ ◯ b) ↔ b ≼{n} a ∧ ✓{n} a.
Proof.
  rewrite dauth_both_frac_validN -cmra_discrete_valid_iff frac_valid'. naive_solver.
Qed. *)

(* Lemma dauth_frag_valid a : ✓ (◯ a) ↔ ✓ a.
Proof. done. Qed.
Lemma dauth_auth_frac_valid q a : ✓ (●{q} a) ↔ ✓ q ∧ ✓ a.
Proof.
  rewrite dauth_valid_eq /=. apply and_iff_compat_l. split.
  - intros H'. apply cmra_valid_validN. intros n.
    by destruct (H' n) as [? [->%(inj to_agree) [??]]].
  - intros. exists a. split; [done|].
    split; by [apply ucmra_unit_leastN|apply cmra_valid_validN].
Qed.
Lemma dauth_auth_valid a : ✓ (● a) ↔ ✓ a.
Proof. rewrite dauth_auth_frac_valid frac_valid'. naive_solver. Qed. *)

(* The reverse direction of the two lemmas below only holds if the camera is
discrete. *)
(* Lemma dauth_both_frac_valid_2 q a b : ✓ q → ✓ a → b ≼ a → ✓ (●{q} a ⋅ ◯ b).
Proof.
  intros Val1 Val2 Incl. rewrite dauth_valid_eq /=. split; [done|].
  intros n. exists a. split; [done|]. rewrite left_id.
  split; by [apply cmra_included_includedN|apply cmra_valid_validN].
Qed.
Lemma dauth_both_valid_2 a b : ✓ a → b ≼ a → ✓ (● a ⋅ ◯ b).
Proof. intros ??. by apply dauth_both_frac_valid_2. Qed. *)

(* Lemma dauth_valid_discrete `{!CmraDiscrete A} x :
  ✓ x ↔ match dauth_auth_proj x with
        | Some (q, ag) => ✓ q ∧ ∃ a, ag ≡ to_agree a ∧ dauth_frag_proj x ≼ a ∧ ✓ a
        | None => ✓ dauth_frag_proj x
        end.
Proof.
  rewrite dauth_valid_eq. destruct x as [[[??]|] ?]; simpl; [|done].
  setoid_rewrite <-cmra_discrete_included_iff.
  setoid_rewrite <-(discrete_iff _ a).
  setoid_rewrite <-cmra_discrete_valid_iff. naive_solver eauto using O.
Qed.
Lemma dauth_both_frac_valid `{!CmraDiscrete A} q a b :
  ✓ (●{q} a ⋅ ◯ b) ↔ ✓ q ∧ b ≼ a ∧ ✓ a.
Proof.
  rewrite dauth_valid_discrete /=. apply and_iff_compat_l.
  setoid_rewrite (left_id _ _ b). split.
  - by intros [?[->%(inj to_agree)]].
  - naive_solver.
Qed.
Lemma dauth_both_valid `{!CmraDiscrete A} a b : ✓ (● a ⋅ ◯ b) ↔ b ≼ a ∧ ✓ a.
Proof. rewrite dauth_both_frac_valid frac_valid'. naive_solver. Qed. *)

Lemma dauth_cmra_mixin : CmraMixin (dauth K A).
Proof.
  apply cmra_total_mixin; try eauto.
  - by intros n x y1 y2 [Hy Hy']; split; simpl; rewrite ?Hy ?Hy'.
  - by intros n y1 y2 [Hy Hy']; split; simpl; rewrite ?Hy ?Hy'.
  - intros n [x a] [y b] [Hx Ha]; simpl in *. admit.
    (* rewrite !dauth_validN_eq.
    destruct Hx as [x y Hx|]; first destruct Hx as [? Hx];
      [destruct x,y|]; intros VI; ofe_subst; auto. *)
  - admit.
    (* intros [[[]|] ]; rewrite /= ?dauth_valid_eq ?dauth_validN_eq /=
      ?cmra_valid_validN; naive_solver. *)
  (* - intros n [[[]|] ]; rewrite !dauth_validN_eq /=;
      naive_solver eauto using dist_S, cmra_includedN_S, cmra_validN_S. *)
  - by split; simpl; rewrite assoc.
  - by split; simpl; rewrite comm.
  - split; simpl. by rewrite left_id. by rewrite cmra_core_l.
  - by split; simpl; rewrite ?cmra_core_idemp.
  - intros ??; rewrite! dauth_included; intros [??].
    admit.
    (* by split; simpl; apply: cmra_core_mono. FIXME: FIXME(Coq #6294): needs new unification *)
  - assert (∀ n (a b1 b2 : A), b1 ⋅ b2 ≼{n} a → b1 ≼{n} a).
    { intros n a b1 b2 <-; apply cmra_includedN_l. }
    admit.
    (* intros n [[[q1 a1]|] b1] [[[q2 a2]|] b2]; rewrite dauth_validN_eq /=;
      try naive_solver eauto using cmra_validN_op_l, cmra_validN_includedN.
    intros [? [a [Eq [? Valid]]]].
    assert (Eq': a1 ≡{n}≡ a2). { by apply agree_op_invN; rewrite Eq. }
    split; [naive_solver eauto using cmra_validN_op_l|].
    exists a. rewrite -Eq -Eq' agree_idemp. naive_solver. *)
  - intros n x y1 y2 ? [??]; simpl in *.
    admit.
    (* destruct (cmra_extend n (dauth_auth_proj x) (dauth_auth_proj y1)
      (dauth_auth_proj y2)) as (ea1&ea2&?&?&?); auto using dauth_auth_proj_validN.
    destruct (cmra_extend n (dauth_frag_proj x) (dauth_frag_proj y1) (dauth_frag_proj y2))
      as (b1&b2&?&?&?); auto using dauth_frag_proj_validN.
    by exists (DAuth ea1 b1), (DAuth ea2 b2). *)
Admitted.
Canonical Structure dauthR := CmraT (dauth K A) dauth_cmra_mixin.

Global Instance dauth_cmra_discrete : CmraDiscrete A → CmraDiscrete dauthR.
Proof.
Admitted.
  (* split; first apply _.
  intros [[[]|] ?]; rewrite dauth_valid_eq dauth_validN_eq /=; auto.
  - setoid_rewrite <-cmra_discrete_included_iff.
    rewrite -cmra_discrete_valid_iff.
    setoid_rewrite <-cmra_discrete_valid_iff.
    setoid_rewrite <-(discrete_iff _ a). tauto.
  - by rewrite -cmra_discrete_valid_iff.
Qed. *)

Instance dauth_empty : Unit (dauth K A) := DAuth ε ε.
Lemma dauth_ucmra_mixin : UcmraMixin (dauth K A).
Proof.
Admitted.
  (* split; simpl.
  - rewrite dauth_valid_eq /=. apply ucmra_unit_valid.
  - by intros x; constructor; rewrite /= left_id.
  - do 2 constructor; [done| apply (core_id_core _)].
Qed. *)
Canonical Structure dauthUR := UcmraT (dauth K A) dauth_ucmra_mixin.

Global Instance dauth_frag_core_id b : CoreId b → CoreId (◯ b).
Proof. do 2 constructor; simpl; auto. apply: core_id_core. Qed.

Lemma dauth_frag_op b b' : ◯ (b ⋅ b') = ◯ b ⋅ ◯ b'.
Proof. done. Qed.
Lemma dauth_frag_mono b b' : b ≼ b' → ◯ b ≼ ◯ b'.
Proof. intros [c ->]. rewrite dauth_frag_op. apply cmra_included_l. Qed.
Lemma dauth_frag_core b : core (◯ b) = ◯ (core b).
Proof. done. Qed.

Lemma dauth_update_frag_singleton (q q' : dfrac) i v :
  q ~~> q' → ◯ {[ i := (q, v) ]} ~~> ◯ {[ i := (q', v) ]}.
Proof.

Global Instance dauth_frag_sep_homomorphism :
  MonoidHomomorphism op op (≡) (@dauth_frag A).
Proof. by split; [split; try apply _|]. Qed.

Lemma big_opL_auth_frag {B} (g : nat → B → A) (l : list B) :
  (◯ [^op list] k↦x ∈ l, g k x) ≡ [^op list] k↦x ∈ l, ◯ (g k x).
Proof. apply (big_opL_commute _). Qed.
Lemma big_opM_auth_frag `{Countable K} {B} (g : K → B → A) (m : gmap K B) :
  (◯ [^op map] k↦x ∈ m, g k x) ≡ [^op map] k↦x ∈ m, ◯ (g k x).
Proof. apply (big_opM_commute _). Qed.
Lemma big_opS_auth_frag `{Countable B} (g : B → A) (X : gset B) :
  (◯ [^op set] x ∈ X, g x) ≡ [^op set] x ∈ X, ◯ (g x).
Proof. apply (big_opS_commute _). Qed.
Lemma big_opMS_auth_frag `{Countable B} (g : B → A) (X : gmultiset B) :
  (◯ [^op mset] x ∈ X, g x) ≡ [^op mset] x ∈ X, ◯ (g x).
Proof. apply (big_opMS_commute _). Qed.

Lemma dauth_both_op q a b : DAuth (Some (q,to_agree a)) b ≡ ●{q} a ⋅ ◯ b.
Proof. by rewrite /op /dauth_op /= left_id. Qed.

Lemma dauth_auth_frac_op p q a : ●{p + q} a ≡ ●{p} a ⋅ ●{q} a.
Proof.
  intros; split; simpl; last by rewrite left_id.
  by rewrite -Some_op -pair_op agree_idemp.
Qed.
Lemma dauth_auth_frac_op_invN n p a q b : ✓{n} (●{p} a ⋅ ●{q} b) → a ≡{n}≡ b.
Proof.
  rewrite /op /dauth_op /= left_id -Some_op -pair_op dauth_validN_eq /=.
  intros (?&?& Eq &?&?). apply (inj to_agree), agree_op_invN. by rewrite Eq.
Qed.
Lemma dauth_auth_frac_op_inv p a q b : ✓ (●{p} a ⋅ ●{q} b) → a ≡ b.
Proof.
  intros ?. apply equiv_dist. intros n.
  by eapply dauth_auth_frac_op_invN, cmra_valid_validN.
Qed.
Lemma dauth_auth_frac_op_invL `{!LeibnizEquiv A} q a p b :
  ✓ (●{p} a ⋅ ●{q} b) → a = b.
Proof. by intros ?%dauth_auth_frac_op_inv%leibniz_equiv. Qed.

(** Internalized properties *)
Lemma dauth_equivI {M} x y :
  x ≡ y ⊣⊢@{uPredI M} dauth_auth_proj x ≡ dauth_auth_proj y ∧ dauth_frag_proj x ≡ dauth_frag_proj y.
Proof. by uPred.unseal. Qed.
Lemma dauth_validI {M} x :
  ✓ x ⊣⊢@{uPredI M} match dauth_auth_proj x with
                    | Some (q, ag) => ✓ q ∧
                      ∃ a, ag ≡ to_agree a ∧ (∃ b, a ≡ dauth_frag_proj x ⋅ b) ∧ ✓ a
                    | None => ✓ dauth_frag_proj x
                    end.
Proof. uPred.unseal. by destruct x as [[[]|]]. Qed.
Lemma dauth_frag_validI {M} (a : A):
  ✓ (◯ a) ⊣⊢@{uPredI M} ✓ a.
Proof. by rewrite dauth_validI. Qed.

Lemma dauth_both_validI {M} q (a b: A) :
  ✓ (●{q} a ⋅ ◯ b) ⊣⊢@{uPredI M} ✓ q ∧ (∃ c, a ≡ b ⋅ c) ∧ ✓ a.
Proof.
  rewrite -dauth_both_op dauth_validI /=. apply bi.and_proper; [done|].
  setoid_rewrite agree_equivI. iSplit.
  - iDestruct 1 as (a') "[#Eq [H V]]". iDestruct "H" as (b') "H".
    iRewrite -"Eq" in "H". iRewrite -"Eq" in "V". auto.
  - iDestruct 1 as "[H V]". iExists a. auto.
Qed.
Lemma dauth_auth_validI {M} q (a b: A) :
  ✓ (●{q} a) ⊣⊢@{uPredI M} ✓ q ∧ ✓ a.
Proof.
  rewrite dauth_validI /=. apply bi.and_proper; [done|].
  setoid_rewrite agree_equivI. iSplit.
  - iDestruct 1 as (a') "[Eq [_ V]]". by iRewrite -"Eq" in "V".
  - iIntros "V". iExists a. iSplit; [done|]. iSplit; [|done]. iExists a.
    by rewrite left_id.
Qed.

Lemma dauth_update a b a' b' :
  (a,b) ~l~> (a',b') → ● a ⋅ ◯ b ~~> ● a' ⋅ ◯ b'.
Proof.
  intros Hup; apply cmra_total_update.
  move=> n [[qa|] bf1] [/= Hq [a0 [Haa0 [[bf2 Ha] Ha0]]]]; do 2 red; simpl in *.
  { by apply (exclusiveN_l _ _) in Hq. }
  split; [done|]. apply (inj to_agree) in Haa0.
  move: Ha; rewrite !left_id -assoc=> Ha.
  destruct (Hup n (Some (bf1 ⋅ bf2))); [by rewrite Haa0..|]; simpl in *.
  exists a'. split_and!; [done| |done].
  exists bf2. by rewrite left_id -assoc.
Qed.

Lemma dauth_update_alloc a a' b' : (a,ε) ~l~> (a',b') → ● a ~~> ● a' ⋅ ◯ b'.
Proof. intros. rewrite -(right_id _ _ (● a)). by apply dauth_update. Qed.

Lemma dauth_update_auth a a' b' : (a,ε) ~l~> (a',b') → ● a ~~> ● a'.
Proof.
  intros. etrans; first exact: dauth_update_alloc.
  exact: cmra_update_op_l.
Qed.

Lemma dauth_update_core_id q a b `{!CoreId b} :
  b ≼ a → ●{q} a ~~> ●{q} a ⋅ ◯ b.
Proof.
  intros Ha%(core_id_extract _ _). apply cmra_total_update.
  move=> n [[[q' ag]|] bf1] [Hq [a0 [Haa0 [Hbf1 Ha0]]]]; do 2 red; simpl in *.
  + split; [done|]. exists a0. split_and!; [done| |done].
    assert (a ≡{n}≡ a0) as Haa0' by (apply to_agree_includedN; by exists ag).
    rewrite -Haa0' Ha Haa0' -assoc (comm _ b) assoc. by apply cmra_monoN_r.
  + split; [done|]. exists a0. split_and!; [done| |done].
    apply (inj to_agree) in Haa0.
    rewrite -Haa0 Ha Haa0 -assoc (comm _ b) assoc. by apply cmra_monoN_r.
Qed.

Lemma dauth_update_dealloc a b a' : (a,b) ~l~> (a',ε) → ● a ⋅ ◯ b ~~> ● a'.
Proof. intros. rewrite -(right_id _ _ (● a')). by apply dauth_update. Qed.

Lemma dauth_local_update (a b0 b1 a' b0' b1': A) :
  (b0, b1) ~l~> (b0', b1') → b0' ≼ a' → ✓ a' →
  (● a ⋅ ◯ b0, ● a ⋅ ◯ b1) ~l~> (● a' ⋅ ◯ b0', ● a' ⋅ ◯ b1').
Proof.
  rewrite !local_update_unital=> Hup ? ? n /=.
    move=> [[[qc ac]|] bc] /dauth_both_validN [Le Val] [] /=.
  - move => Ha. exfalso. move : Ha. rewrite right_id -Some_op -pair_op frac_op'.
    move => /Some_dist_inj [/= Eq _].
    apply (Qp_not_plus_q_ge_1 qc). by rewrite -Eq.
  - move => _. rewrite !left_id=> ?.
    destruct (Hup n bc) as [Hval' Heq]; eauto using cmra_validN_includedN.
    rewrite -!dauth_both_op dauth_validN_eq /=.
    split_and!; [done| |done]. exists a'.
    split_and!; [done|by apply cmra_included_includedN|by apply cmra_valid_validN].
Qed.
End cmra.

Arguments dauthR : clear implicits.
Arguments dauthUR : clear implicits.

(* Proof mode class instances *)
Instance is_op_auth_frag {A : ucmraT} (a b1 b2 : A) :
  IsOp a b1 b2 → IsOp' (◯ a) (◯ b1) (◯ b2).
Proof. done. Qed.
Instance is_op_auth_auth_frac {A : ucmraT} (q q1 q2 : frac) (a : A) :
  IsOp q q1 q2 → IsOp' (●{q} a) (●{q1} a) (●{q2} a).
Proof. rewrite /IsOp' /IsOp => ->. by rewrite -dauth_auth_frac_op. Qed.

(* Functor *)
Definition dauth_map {A B} (f : A → B) (x : dauth A) : dauth B :=
  DAuth (prod_map id (agree_map f) <$> dauth_auth_proj x) (f (dauth_frag_proj x)).
Lemma dauth_map_id {A} (x : dauth A) : dauth_map id x = x.
Proof. destruct x as [[[]|] ]; by rewrite // /dauth_map /= agree_map_id. Qed.
Lemma dauth_map_compose {A B C} (f : A → B) (g : B → C) (x : dauth A) :
  dauth_map (g ∘ f) x = dauth_map g (dauth_map f x).
Proof. destruct x as [[[]|] ];  by rewrite // /dauth_map /= agree_map_compose. Qed.
Lemma dauth_map_ext {A B : ofeT} (f g : A → B) `{!NonExpansive f} x :
  (∀ x, f x ≡ g x) → dauth_map f x ≡ dauth_map g x.
Proof.
  constructor; simpl; auto.
  apply option_fmap_equiv_ext=> a; by rewrite /prod_map /= agree_map_ext.
Qed.
Instance dauth_map_ne {A B : ofeT} (f : A -> B) `{Hf : !NonExpansive f} :
  NonExpansive (dauth_map f).
Proof.
  intros n [??] [??] [??]; split; simpl in *; [|by apply Hf].
  apply option_fmap_ne; [|done]=> x y ?. apply prod_map_ne; [done| |done].
  by apply agree_map_ne.
Qed.
Instance dauth_map_cmra_morphism {A B : ucmraT} (f : A → B) :
  CmraMorphism f → CmraMorphism (dauth_map f).
Proof.
  split; try apply _.
  - intros n [[[??]|] b]; rewrite !dauth_validN_eq;
      [|naive_solver eauto using cmra_morphism_monotoneN, cmra_morphism_validN].
    intros [? [a' [Eq [? ?]]]]. split; [done|]. exists (f a').
    rewrite -agree_map_to_agree -Eq.
    naive_solver eauto using cmra_morphism_monotoneN, cmra_morphism_validN.
  - intros [??]. apply Some_proper; rewrite /dauth_map /=.
    by f_equiv; rewrite /= cmra_morphism_core.
  - intros [[[??]|]?] [[[??]|]?]; try apply DAuth_proper=>//=; by rewrite cmra_morphism_op.
Qed.
Definition dauthO_map {A B} (f : A -n> B) : dauthO A -n> dauthO B :=
  OfeMor (dauth_map f).
Lemma dauthO_map_ne A B : NonExpansive (@dauthO_map A B).
Proof. intros n f f' Hf [[[]|] ]; repeat constructor; try naive_solver;
  apply agreeO_map_ne; auto. Qed.

Program Definition dauthRF (F : urFunctor) : rFunctor := {|
  rFunctor_car A _ B _ := dauthR (urFunctor_car F A B);
  rFunctor_map A1 _ A2 _ B1 _ B2 _ fg := dauthO_map (urFunctor_map F fg)
|}.
Next Obligation.
  by intros F A1 ? A2 ? B1 ? B2 ? n f g Hfg; apply dauthO_map_ne, urFunctor_map_ne.
Qed.
Next Obligation.
  intros F A ? B ? x. rewrite /= -{2}(dauth_map_id x).
  apply (dauth_map_ext _ _)=>y; apply urFunctor_map_id.
Qed.
Next Obligation.
  intros F A1 ? A2 ? A3 ? B1 ? B2 ? B3 ? f g f' g' x. rewrite /= -dauth_map_compose.
  apply (dauth_map_ext _ _)=>y; apply urFunctor_map_compose.
Qed.

Instance dauthRF_contractive F :
  urFunctorContractive F → rFunctorContractive (dauthRF F).
Proof.
  by intros ? A1 ? A2 ? B1 ? B2 ? n f g Hfg; apply dauthO_map_ne, urFunctor_map_contractive.
Qed.

Program Definition dauthURF (F : urFunctor) : urFunctor := {|
  urFunctor_car A _ B _ := dauthUR (urFunctor_car F A B);
  urFunctor_map A1 _ A2 _ B1 _ B2 _ fg := dauthO_map (urFunctor_map F fg)
|}.
Next Obligation.
  by intros F A1 ? A2 ? B1 ? B2 ? n f g Hfg; apply dauthO_map_ne, urFunctor_map_ne.
Qed.
Next Obligation.
  intros F A ? B ? x. rewrite /= -{2}(dauth_map_id x).
  apply (dauth_map_ext _ _)=>y; apply urFunctor_map_id.
Qed.
Next Obligation.
  intros F A1 ? A2 ? A3 ? B1 ? B2 ? B3 ? f g f' g' x. rewrite /= -dauth_map_compose.
  apply (dauth_map_ext _ _)=>y; apply urFunctor_map_compose.
Qed.

Instance dauthURF_contractive F :
  urFunctorContractive F → urFunctorContractive (dauthURF F).
Proof.
  by intros ? A1 ? A2 ? B1 ? B2 ? n f g Hfg; apply dauthO_map_ne, urFunctor_map_contractive.
Qed.
