From reloc Require Import reloc lib.lock.
From msqueue Require Import common extra_rules CG_queue MS_queue_no_snapshot.
From iris.algebra Require Import numbers csum excl auth list gmap gset.
From iris_string_ident Require Import ltac2_string_ident.
From iris.bi.lib Require Export fixpoint.
From iris_string_ident Require Import ltac2_string_ident.

Definition queueN : namespace := nroot .@ "queue".

Definition setUR : ucmraT := authUR (gsetUR loc).

Class queueG Σ := {
  queue_setUR :> inG Σ setUR;
}.

Definition queueΣ := #[ GFunctor setUR ].
Instance subG_lockPoolΣ {Σ} : subG queueΣ Σ → queueG Σ.
Proof. solve_inG. Qed.

Section queue_refinement.
  Context `{relocG Σ, queueG Σ, !lockG Σ}.

  (* Representation predicate for the MS-queue. *)
  Fixpoint isMSQueue (ℓt ℓtOut ℓp ℓpOut : loc) (xs : list val) : iProp Σ :=
    match xs with
    | nil => ⌜ℓp = ℓt⌝ ∗ ⌜ℓpOut = ℓtOut⌝
    | x :: xs' =>
      (∃ (ℓn ℓnOut : loc),
          ℓpOut ↦ₛ{1/2} #ℓn ∗ ℓn ↦ₛ{1/2} CONSV (SOMEV x) #ℓnOut
        ∗ isMSQueue ℓt ℓtOut ℓn ℓnOut xs')
    end.

  Definition N := nroot.@"lock".

  (* Predicate expressing that ℓtoS points to a queue with the values xs *)
  Definition msQueueInv (ℓtoS ℓtoT : loc) (xsₛ : list val) : iProp Σ :=
    (∃ (ℓs ℓsOut ℓt ℓtOut ℓnil : loc) vs q,
          ℓtoS ↦ₛ #ℓs
        ∗ ℓtoT ↦ₛ #ℓt
        ∗ ℓs ↦ₛ{q} SOMEV (vs, #ℓsOut)
        ∗ ℓtOut ↦ₛ #ℓnil ∗ ℓnil ↦ₛ NONEV
        ∗ isMSQueue ℓt ℓtOut ℓs ℓsOut xsₛ).

  Definition cgQueueInv ℓs (xsᵢ : list val) : iProp Σ := 
    (ℓs ↦ isCGQueue_go xsᵢ)%I.

  Definition queueInv (τi : lrel Σ) ℓtoS ℓtoT ℓs : iProp Σ :=
    (∃ xsᵢ xsₛ,
        cgQueueInv ℓs xsᵢ
      ∗ msQueueInv ℓtoS ℓtoT xsₛ
      ∗ [∗ list] xᵢ ; xₛ ∈ xsᵢ ; xsₛ, τi xᵢ xₛ)%I.

  Definition ctx := Datatypes.list (ectxi_language.ectx_item heap_ectxi_lang).

  Local Hint Resolve tpool_lookup : core.
  Local Hint Resolve tpool_lookup_Some : core.
  Local Hint Resolve to_tpool_insert : core.
  Local Hint Resolve to_tpool_insert' : core.
  Local Hint Resolve tpool_singleton_included : core.
  
  Lemma step_resolve_load E j K l q v (p : proph_id) w :
    nclose specN ⊆ E →
    spec_ctx ∗ l ↦ₛ{q} v ∗ j ⤇ fill K (Resolve (Load #l) #p (of_val w)) ={E}=∗
    j ⤇ fill K v ∗ l ↦ₛ{q} v.
  Proof.
    iIntros (?) "(#Hinv & Hl & Hj)".
    rewrite /spec_ctx tpool_mapsto_eq /tpool_mapsto_def /=.
    iDestruct "Hinv" as (ρ) "Hinv".
    iInv specN as (tp σ) ">[Hown %]" "Hclose".
    iDestruct (own_valid_2 with "Hown Hj")
      as %[[?%tpool_singleton_included' _]%prod_included _]%auth_both_valid.
    iMod (own_update_2 with "Hown Hj") as "[Hown Hj]".
    { by eapply auth_update, prod_local_update_1,
         singleton_local_update, (exclusive_local_update _ (Excl (fill K v))). }
    (* Get the points-to predicate. *)
    rewrite heapS_mapsto_eq /heapS_mapsto_def /=.
    iDestruct (own_valid_2 with "Hown Hl")
      as %[[? ?%gen_heap_singleton_included]%prod_included ?]%auth_both_valid.
    iFrame. iApply "Hclose". iNext.
    iExists (<[j:=fill K v]> tp), σ.
    rewrite to_tpool_insert'; last eauto. iFrame. iPureIntro.
    eapply rtc_r, step_insert_no_fork; eauto.
    econstructor. econstructor. auto.
  Qed.

  Lemma refines_resolve_load_r E K l v t (p : proph_id) w A
    (Hmasked : nclose specN ⊆ E) :
    l ↦ₛ v -∗ 
    (l ↦ₛ v -∗ REL t << fill K (of_val v) @ E : A)%I
    -∗ REL t << fill K (Resolve (! #l) #p (of_val w)) @ E : A.
  Proof.
    iIntros "Hl Hlog".
    pose (Φ := (fun w => ⌜w = v⌝ ∗ l ↦ₛ v)%I).
    iApply (refines_step_r Φ with "[Hl]"); simpl; auto.
    { cbv[Φ].
      iIntros (j K') "#Hs Hj /=".
      iMod (step_resolve_load with "[$Hs $Hl $Hj]") as "[Hj Hl]". done.
      iModIntro. iExists _. iFrame. eauto. }
    iIntros (v') "He'". iDestruct "He'" as "[-> Hl]".
    by iApply "Hlog".
  Qed.

  Lemma refines_CG_enqueue_body_l E t (K : Datatypes.list (ectxi_language.ectx_item heap_ectxi_lang)) A x xs :
    nclose relocN ⊆ E →
    (REL fill K (isCGQueue_go (xs ++ [x])) << t @ E : A) -∗
    (REL fill K (CG_enqueue_go x (isCGQueue_go xs)) << t @ E : A).
  Proof.
    iIntros (HNE) "H".
    iInduction xs as [|x' xs'] "IH" forall (K). simpl.
    - rel_rec_l. rel_pures_l. rel_pures_l.
      iApply "H".
    - simpl.
      rel_rec_l. rel_pures_l. rel_pures_l. rel_apply_l ("IH").
      rel_pures_l. done.
  Qed.

  Theorem changed_last_pointer (ℓt ℓtOut ℓs ℓsOut ℓn ℓtoTail ℓnil' : loc) xs x :
    isMSQueue ℓt ℓtOut ℓs ℓsOut xs -∗
    ℓtOut ↦ₛ #ℓn -∗
    ℓn ↦ₛ CONSV (InjRV x) #ℓtoTail -∗
    ℓtoTail ↦ₛ #ℓnil' -∗
    ℓnil' ↦ₛ NILV -∗
    (isMSQueue ℓn ℓtoTail ℓs ℓsOut (xs ++ [x]) ∗
     ℓn ↦ₛ{1/2} CONSV (InjRV x) #ℓtoTail ∗
     ℓtoTail ↦ₛ #ℓnil' ∗
     ℓnil' ↦ₛ NILV).
  Proof.
    iIntros "nodeList tOutPts nodePts toTailPts nilPts".
    iInduction xs as [|x' xs'] "IH" forall (ℓs ℓsOut).
    - iDestruct "nodeList" as %[<- <-].
      iDestruct "nodePts" as "[??]".
      iFrame. simpl.
      iExists _, _. iDestruct "tOutPts" as "[??]". iFrame. auto.
    - iDestruct "nodeList" as (??) "(? & ? & nodeList)".
      fold isMSQueue.
      iDestruct ("IH" with "nodeList tOutPts nodePts toTailPts nilPts") as "(nodePts&?&?)".
      iFrame. iExistsN. iFrame.
  Qed.

  Lemma get_tail_pointer (ℓt ℓtOut ℓs ℓsOut : loc) xs v q :
    ℓs ↦ₛ{q} CONSV v #ℓsOut -∗
    isMSQueue ℓt ℓtOut ℓs ℓsOut xs -∗
    (∃ v' q', ℓt ↦ₛ{q'} CONSV v' #ℓtOut ∗ (ℓt ↦ₛ{q'} (CONSV v' #ℓtOut) -∗ ℓs ↦ₛ{q} CONSV v #ℓsOut∗ isMSQueue ℓt ℓtOut ℓs ℓsOut xs)).
  Proof.
    iIntros "sPts isMS".
    iInduction xs as [|x xs'] "IH" forall (ℓs ℓsOut q v).
    - iDestruct "isMS" as "[-> ->]".
      iExists _, _. iFrame. iIntros "tPts". iFrame. auto.
    - simpl.
      iDestruct "isMS" as (??) "(sOutPts & nPts & isMS)".
      iDestruct ("IH" with "nPts isMS") as (??) "(?& Rest)".
      iExistsN. iFrame. iIntros "tPts".
      iExistsN. iFrame. iApply "Rest". iFrame.
  Qed.

  Lemma t_s_neq (ℓs ℓsOut ℓt ℓtOut ℓnil : loc) x xs q vs :
    ℓs ↦ₛ{q} CONSV vs #ℓsOut -∗
    ℓtOut ↦ₛ #ℓnil -∗
    ℓnil ↦ₛ NILV -∗
    isMSQueue ℓt ℓtOut ℓs ℓsOut (x :: xs) -∗
    ⌜ℓs ≠ ℓt⌝.
  Proof.
    iIntros "sPts tOutPts nilPts nodeList".
    simpl.
    iDestruct "nodeList" as (??) "(a & [nPts nPts'] & nodeList)".
    iDestruct (get_tail_pointer with "nPts nodeList") as (??) "[tPts Rest]".
    iIntros (->).
    iDestruct (mapstoS_agree with "sPts tPts") as %[= -> ->].
    iDestruct (mapstoS_agree with "a tOutPts") as %[= ->].
    iDestruct (mapstoS_agree with "nilPts nPts'") as %[= ->].
  Qed.

  Lemma queue_refinement :
    ⊢ REL CG_queue << MS_queue_no_snapshot : ∀ A, (() → () + A) * (A → ()).
  Proof.
    iApply refines_forall. iModIntro. iIntros (τi).
    (* Step through the initialization of the specification. *)
    rel_alloc_r nil as "nilPts".
    rel_alloc_r tail as "tailPts".
    rel_alloc_r ℓsent as "[sentPts sentPts']".
    rel_alloc_r ℓtoT as "toTPts".
    rel_alloc_r ℓtoS as "toSPts".
    rel_pures_r.
    (* Step through the initialization of the implementation. *)
    rel_alloc_l ℓs as "sPts".
    rel_pures_l.
    rel_apply_l (refines_newlock_l N (queueInv τi ℓtoS ℓtoT ℓs) with "[-]").
    { iExists [], []. simpl. iFrame. iSplitL; last done.
      iExistsN. iFrame "toSPts". iFrame. done. }
    iNext. iIntros (lk γ) "#Hlk".
    rel_pure_l.
    rel_pure_l.
    rel_pure_l.
    rel_pure_l.

    iApply refines_pair.
    - (* dequeue *)
      rewrite /MS_dequeue_no_snapshot /CG_dequeue.
      rel_pures_r.
      iApply refines_arrow_val. iModIntro.
      iIntros (? ? [-> ->]). rel_pures_l. clear.
      rel_apply_l (refines_acquire_l with "Hlk").
      iNext. iIntros "Htok".
      iDestruct 1 as ([|xᵢ xsᵢ'] xsₛ) "(sPts & Hms & Hlink)".
      * rel_pures_l.
        destruct xsₛ as [|x' xsₛ'].
        2: { iDestruct (big_sepL2_length with "Hlink") as %[=]. }
        iDestruct "Hms" as (???????) "(toSPts & toTPts & sentPts & tOutPts & nilPts & [-> ->])".
        rel_load_r.
        rel_load_r.
        rel_load_r.
        rel_rec_r.
        rel_pures_r.
        rel_load_r.
        rel_pures_r.
        rel_pures_r.
        rewrite bool_decide_true; last done.
        rel_pures_r.
        rel_pures_r.
        rel_load_r.
        rel_pures_r.

        rel_rec_l.
        rel_load_l.
        rel_pures_l.
        rel_apply_l (refines_release_l with "Hlk Htok [-]").
        { iExistsFrame. }
        iNext.
        rel_pures_l.
        rel_values.
        iExistsN. auto.
      * rel_pures_l.
        rel_rec_l.
        rel_load_l.
        rel_pures_l.
        rel_store_l.
        rel_pures_l.
        destruct xsₛ as [|x' xsₛ'].
        { iDestruct (big_sepL2_length with "Hlink") as %[=]. }
        iDestruct "Hms" as (???????) "(toSPts & toTPts & sentPts & tOutPts & nilPts & nodeList)".
        iDestruct (t_s_neq with "sentPts [$] nilPts nodeList") as %Hneq.
        iDestruct "nodeList" as (??) "(soutPts & nPts & nodeList)".
        rel_load_r.
        rel_load_r.
        rel_load_r.
        rel_rec_r.
        rel_pures_r.
        simpl.
        rel_load_r.
        rel_pures_r.
        rewrite bool_decide_false. 2: { intros [=]. done. }
        rel_pures_r.
        rel_cmpxchg_suc_r.
        rel_pures_r.
        rel_load_r.
        rel_rec_r.
        rel_pures_r.
        rel_rec_r.
        rel_pures_r.
        iDestruct "Hlink" as "[Hτi Hlink]".
        rel_apply_l (refines_release_l with "Hlk Htok [-Hτi]").
        { iExists (xsᵢ'), (xsₛ'). iFrame.
          iFrame. iExistsN. iFrame "toTPts toSPts ∗". }
        iNext. rel_pures_l. rel_values. iExistsN. auto.
    - rewrite /MS_enqueue_no_snapshot /CG_enqueue.
      rel_pures_l.
      rel_pures_r.
      iApply refines_arrow_val. iModIntro.
      iIntros (x x') "#Hrel". rel_pures_l.
      rel_apply_l (refines_acquire_l with "Hlk").
      iNext. iIntros "Htok".
      iDestruct 1 as (xsₛ xsᵢ) "(sPts & Hms & Hlink)".
      iDestruct "Hms" as (???????) "(toSPts & toTPts & sentPts & tOutPts & nilPts & nodeList)".
      rel_load_l.
      rel_apply_l refines_CG_enqueue_body_l; auto.
      rel_store_l.
      rel_pures_l.
      rel_pures_r.
      rel_alloc_r ℓnil' as "nilPts'".
      rel_alloc_r ℓtoTail as "tailPts".
      rel_alloc_r ℓnode as "nodePts".
      rel_load_r.
      rel_pures_r.
      iDestruct (get_tail_pointer with "sentPts nodeList") as (??) "[tPts Rest]".
      rel_load_r.
      iDestruct ("Rest" with "tPts") as "[sentPts nodeList]".
      rel_rec_r.
      rel_pures_r.
      rel_load_r.
      rel_load_r.
      rel_pures_r.
      rel_pures_r.
      rel_cmpxchg_suc_r.
      rel_pures_r.
      rel_cmpxchg_suc_r.
      rel_pures_r.
      iDestruct (changed_last_pointer with "nodeList tOutPts nodePts tailPts nilPts'")
        as "(nodeList & nodePts & tailPts&nilPts')".
      rel_apply_l (refines_release_l with "Hlk Htok [-]").
      { iExists (xsₛ ++ [x]), (xsᵢ ++ [x']). iFrame.
        iSplitL. iExistsN. iFrame "toSPts toTPts". iFrame.
        by iApply big_sepL2_singleton. }
      iNext.
      rel_values.
  Qed.

End queue_refinement.

Open Scope nat.
Theorem CG_ctx_refines_MS_no_snapshot :
  ∅ ⊨ CG_queue ≤ctx≤ MS_queue_no_snapshot :
    ∀: (TUnit → TUnit + TVar 0) * (TVar 0 → TUnit).
Proof.
  pose (Σ := #[relocΣ; queueΣ; lockΣ]).
  eapply (refines_sound Σ).
  iIntros (? Δ).
  iApply queue_refinement.
Qed.
