From reloc Require Import reloc lib.lock.
From msqueue Require Import common extra_rules CG_queue MS_queue MS_queue_no_snapshot.
From iris.algebra Require Import numbers csum excl auth list gmap gset.
From iris_string_ident Require Import ltac2_string_ident.
From iris.bi.lib Require Export fixpoint.
From iris_string_ident Require Import ltac2_string_ident.

Definition queueN : namespace := nroot .@ "queue".

Definition setUR : ucmraT := authUR (gsetUR loc).

Class queueG Σ := {
  queue_setUR :> inG Σ setUR;
}.

Definition queueΣ := #[ GFunctor setUR ].
Instance subG_lockPoolΣ {Σ} : subG queueΣ Σ → queueG Σ.
Proof. solve_inG. Qed.

Section queue_refinement.
  Context `{relocG Σ, queueG Σ}.

  Lemma refines_load_alt K E l t A :
    (|={⊤,E}=> ∃ v' q,
      ▷(l ↦{q} v') ∗
      ▷(l ↦{q} v' -∗ (REL fill K (of_val v') << t @ E : A)))%I
    -∗ REL fill K (! #l) << t : A.
  Proof.
    iIntros "Hlog".
    iApply refines_atomic_l; auto.
    iMod "Hlog" as (v' q) "[Hl Hlog]". iModIntro.
    iApply (wp_load with "Hl"); auto.
  Qed.

  Tactic Notation "rel_load_l_atomic" := rel_apply_l refines_load_alt.

  Definition isNode ℓn x (ℓnOut : loc) : iProp Σ := ℓn ↦□ SOMEV (x, #ℓnOut).

  (* Length indexed reachable *)
  Fixpoint reachable_l (n : nat) ℓn ℓm : iProp Σ :=
    ∃ x (ℓnOut : loc), ℓn ↦□ CONSV x #ℓnOut ∗
      (match n with
      | O => ⌜ℓn = ℓm⌝
      | S n' => (∃ (ℓp : loc), ℓnOut ↦□ #ℓp ∗ reachable_l n' ℓp ℓm)
      end).

  Definition reachable ℓn ℓm : iProp Σ := ∃ n, reachable_l n ℓn ℓm.

  Notation "a ~r~> b" := (reachable a b) (at level 20, format "a  ~r~>  b").

  Lemma reachable_refl x (ℓm ℓmOut : loc) : ℓm ↦□ CONSV x #ℓmOut -∗ ℓm ~r~> ℓm.
  Proof. iIntros "p". iExists 0. iExistsFrame. Qed.

  Instance reachable_persistent a b: Persistent (a ~r~> b).
  Proof.
    rewrite /Persistent.
    iDestruct 1 as (n) "R". iInduction n as [|n] "IH" forall (a).
    - iDestruct "R" as "#R". iModIntro. by iExists 0.
    - iDestruct "R" as (??) "[#? U]". iDestruct "U" as (?) "[#P N]".
      iDestruct ("IH" with "N") as (n') "#HI". iModIntro. iExists (S n'). iExistsFrame.
  Qed.

  Lemma reachable_trans a b c : reachable a b -∗ reachable b c -∗ reachable a c.
  Proof.
    iDestruct 1 as (n) "Ha".
    iInduction n as [|n'] "IH" forall (a); simpl.
    - iDestruct "Ha" as (??) "[? ->]". auto.
    - iIntros "bRc".
      iDestruct "Ha" as (??) "[aPts Hb]".
      iDestruct "Hb" as (?) "[pts R]".
      iDestruct ("IH" with "R bRc") as (n) "Ho".
      iExists (S n). iExistsN. iFrame. iExistsN. iFrame.
  Qed.

  Definition reachable_pre (Φ : loc -d> loc -d> iPropO Σ) : loc -d> loc -d> iPropO Σ :=
    (λ ℓn ℓm, ∃ x (ℓnOut : loc), ℓn ↦□ CONSV x #ℓnOut ∗
                  (⌜ℓn = ℓm⌝ ∨ (∃ (ℓp : loc), ℓnOut ↦□ #ℓp ∗ Φ ℓp ℓm)))%I.

  Lemma reachable_unfold ℓn ℓm : ℓn ~r~> ℓm ⊢ reachable_pre reachable ℓn ℓm.
  Proof.
    iDestruct 1 as ([|n] nOut x) "[nPts H]".
    * iExistsN. iFrame.
    * iExistsN.
      iDestruct "H" as (p) "[nOutPts HI]".
      iFrame. iRight. iExistsN. iFrame. iExists n. iFrame.
  Qed.

  Lemma reachable_is_node (ℓn ℓm : loc) : ℓn ~r~> ℓm -∗ ∃ x ℓnOut, isNode ℓn x ℓnOut.
  Proof. iDestruct 1 as ([|n] nOut x) "[nPts _]"; iExists _, _; iFrame. Qed.

  Lemma reachable_last_is_node (ℓn ℓm : loc) : ℓn ~r~> ℓm -∗ ∃ ℓmOut x, isNode ℓm x ℓmOut.
  Proof.
    iDestruct 1 as (n) "R". iInduction (n) as [|n] "IH" forall (ℓn).
    - iDestruct "R" as (??) "[? ->]". iExists _, _. iFrame.
    - iDestruct "R" as (??) "[_ N]". iDestruct "N" as (?) "[_ N]". iApply ("IH" with "N").
  Qed.

  (* If ℓm is reachable from ℓn and ℓn has NIL as its successor then the two
     nodes are equal. *)
  Lemma reachable_nil_eq (ℓn ℓm ℓnOut ℓp : loc) q x :
    ℓn ~r~> ℓm -∗ ℓn ↦□ CONSV x #ℓnOut -∗ ℓnOut ↦{q} #ℓp -∗ ℓp ↦□ NILV -∗ ⌜ℓn = ℓm⌝.
  Proof.
    iDestruct 1 as ([|n]) "Hi"; iDestruct "Hi" as (??) "[nPts HI]"; iIntros "nPts' nOutPts pPts".
    - iAssumption.
    - iDestruct "HI" as (?) "[pts Hn]".
      iDestruct (mapsto_agree_persistent_2 with "nPts nPts'") as %[= -> ->].
      iDestruct (mapsto_agree_persistent with "pts nOutPts") as %[= <-].
      destruct n; iDestruct "Hn" as (??) "[pPts' _]";
      iDestruct (mapsto_agree_persistent_2 with "pPts pPts'") as %[=].
  Qed.

  Lemma reachable_succ (ℓn ℓm ℓnOut ℓmOut : loc) v v' :
    ℓn ↦□ CONSV v #ℓnOut -∗ ℓnOut ↦□ #ℓm -∗ ℓm ↦□ CONSV v' #ℓmOut -∗ ℓn ~r~> ℓm.
  Proof.
    iIntros "???". iExists 1. iExistsFrame.
  Qed.

  (* If ℓm is the successor of ℓn then ℓn ~r~> ℓo for a distinct ℓo implies ℓm ~r~> ℓo. *)
  Lemma reachable_transp (ℓn ℓm ℓnOut ℓo ℓmOut : loc) v v' :
    ℓn ≠ ℓo →
    ℓn ~r~> ℓo
    -∗ ℓn ↦□ CONSV v #ℓnOut -∗ ℓnOut ↦□ #ℓm -∗ ℓm ↦□ CONSV v' #ℓmOut
    -∗ ℓm ~r~> ℓo.
  Proof.
    rewrite reachable_unfold.
    iDestruct 1 as (x ℓm') "[nPts [%|Hi]]"; first done.
    iDestruct "Hi" as (?) "[nOutPts reach]".
    iIntros "nPts' nOutPts' c".
    iDestruct (mapsto_agree_persistent_2 with "nPts nPts'") as %[= -> ->].
    iDestruct (mapsto_agree_persistent_2 with "nOutPts nOutPts'") as %[= ->].
    done.
  Qed.

  (* Fixpoint reachable *)

  (* Locate "-d>". *)
  Definition reachable_pre' (Φ : (loc * loc) -d> iPropO Σ) : (loc * loc) -d> iPropO Σ :=
    (λ '(ℓn, ℓm),
        ∃ x (ℓnOut : loc), ℓn ↦□ CONSV x #ℓnOut ∗
            (⌜ℓn = ℓm⌝ ∨ (∃ (ℓp : loc), ℓnOut ↦□ #ℓp ∗ Φ (ℓp, ℓm))))%I.

  Instance reachable_pre_ne : NonExpansive2 reachable_pre'.
  Proof.
    intros ? ? ? ? (?, ?) (?, ?) [->%leibniz_equiv ->%leibniz_equiv]%pair_equiv_inj.
    rewrite /reachable_pre'.
    do 9 f_equiv. apply H1.
  Qed.

  Definition reachable' : (loc * loc) → iPropI Σ := bi_least_fixpoint (reachable_pre').

  Notation "a ~r'~> b" := (reachable' (a, b)) (at level 20, format "a  ~r'~>  b").

  Lemma reachable_pre_mono (Φ Ψ : (loc * loc) → iProp Σ) :
    ⊢ <pers> (∀ t, Φ t -∗ Ψ t) →
      ∀ t, reachable_pre' Φ t -∗ reachable_pre' Ψ t.
  Proof.
    iIntros "#H". iIntros ((ℓn, ℓm)). rewrite /reachable_pre'.
    iDestruct 1 as (x ?) "[nPts [?|Hl]]"; iExistsN; iFrame.
    iDestruct "Hl" as (?) "[pts R]". iRight. iExistsN. iFrame. iApply "H". iFrame.
  Qed.

  Local Instance reachable_pre_mono' : BiMonoPred reachable_pre'.
  Proof.
    constructor; first apply reachable_pre_mono.
    intros wp Hwp n t1 t2 ?%(discrete_iff _ _)%leibniz_equiv; solve_proper.
  Qed.

  Lemma reachable_unfold' ℓn ℓm : reachable' (ℓn, ℓm) ⊣⊢ reachable_pre' reachable' (ℓn, ℓm).
  Proof. by rewrite /reachable' least_fixpoint_unfold. Qed.

  (*
    Lemma least_fixpoint_ind (Φ : A → PROP) `{!NonExpansive Φ} :
      □ (∀ y, F Φ y -∗ Φ y) -∗ ∀ x, bi_least_fixpoint F x -∗ Φ x.
  *)

  (* Definition RΦ o m : loc -d> iPropO Σ := (λ n, o ~r'~> m -∗ n ~r'~> m)%I. *)

  Lemma reachable_trans' n o m : n ~r'~> o -∗ o ~r'~> m -∗ n ~r'~> m.
  Proof.
    iApply (least_fixpoint_ind reachable_pre' (λ '(n, o), o ~r'~> m -∗ n ~r'~> m)
      with "[]")%I.
    iModIntro. rewrite /reachable_pre'. simpl. iIntros ((n', o')).
    iDestruct 1 as (? ?) "[pts [%eq|He]] /=".
    - subst. auto.
    - iIntros "oRm".
      iDestruct "He" as (?) "[He IH]".
      iDestruct ("IH" with "oRm") as "Rm".
      iEval (rewrite reachable_unfold').
      iExistsN. iFrame. iRight. iExistsN. iFrame.
    Unshelve. 
    iIntros (? (?, ?) (?,?)). rewrite reachable_unfold'. intros I.
    apply pair_equiv_inj in I. destruct I as [->%leibniz_equiv ->%leibniz_equiv].
    simpl. f_equiv. rewrite reachable_unfold'. solve_proper.
  Qed.

  Notation "a -r-> b" := (own b (◯ {[ a ]})) (at level 20).

  Definition tie γt ℓt : iProp Σ := ∃ s, own γt (● s) ∗ ([∗ set] ℓn ∈ s, ℓn ~r~> ℓt).

  Definition node_info γn ℓn x ℓnOut : iProp Σ := isNode ℓn x ℓnOut ∗ tie γn ℓn.

  (* Representation predicate for the MS-queue. *)
  Fixpoint isMSQueue (ℓtoNil ℓtoN : loc) (xs : list val) : iProp Σ :=
    match xs with
    | nil => ⌜ℓtoNil = ℓtoN⌝
    | x :: xs' =>
      (∃ (ℓn ℓnOut : loc),
          ℓtoN ↦□ #ℓn ∗ ℓn ↦□ CONSV (SOMEV x) #ℓnOut
        ∗ isMSQueue ℓtoNil ℓnOut xs')
    end.

  Instance isMSQueue_persistent ℓtoNil ℓtoN xs : Persistent (isMSQueue ℓtoNil ℓtoN xs).
  Proof.
    rewrite /Persistent. 
    iInduction xs as [|n'] "IH" forall (ℓtoN); first auto.
    iDestruct 1 as (??) "(#? & #? & Hr)". iExistsN. iFrame "#". by iApply "IH".
  Qed.

  (* Predicate expressing that ℓtoS points to a queue with the values xs *)
  Definition msQueueInv γs γt γl (ℓtoS ℓtoT : loc) (xsᵢ : list val) : iProp Σ :=
    (∃ (ℓs ℓsOut ℓt ℓtOut ℓl ℓlOut ℓnil : loc) vs vt vl,
          ℓtoS ↦ #ℓs
        ∗ ℓtoT ↦ #ℓt
        ∗ (isMSQueue ℓlOut ℓsOut xsᵢ ∗ ℓlOut ↦ #ℓnil ∗ ℓnil ↦□ NONEV)
        ∗ node_info γs ℓs vs ℓsOut
        ∗ node_info γt ℓt vt ℓtOut
        ∗ node_info γl ℓl vl ℓlOut
        ∗ ℓs -r-> γt (* The tail is reachable from the sentinel. *)
        ∗ ℓt -r-> γl (* The last node is reachable from the tail. *)).

  Definition queueInv γs γt γl (τi : lrel Σ) ℓtoS ℓtoT ℓs lk : iProp Σ :=
    (∃ xsᵢ xsₛ,
        msQueueInv γs γt γl ℓtoS ℓtoT xsᵢ
      ∗ cgQueueInv ℓs lk xsₛ
      ∗ [∗ list] xᵢ ; xₛ ∈ xsᵢ ; xsₛ, τi xᵢ xₛ)%I.

  Lemma is_ms_queue_eq (ℓnOut ℓnil : loc) xs : isMSQueue ℓnOut ℓnOut xs  -∗ ℓnOut ↦ #ℓnil -∗ ℓnil ↦□ NONEV -∗ ⌜xs = []⌝.
  Proof.
    iIntros "nodeList nOutPts nilPts".
    destruct xs as [|xs]; first done.
    iDestruct "nodeList" as (ℓn ℓnOut') "(nOutPts' & nPts & nodeList)".
    iDestruct (mapsto_agree_persistent with "nOutPts' nOutPts") as %[= ->].
    iDestruct (mapsto_agree_persistent_2 with "nilPts nPts") as %[= ->].
  Qed.

  Lemma lookup_reach_set_map s ℓn ℓt :
    ℓn ∈ s → ([∗ set] ℓn ∈ s, ℓn ~r~> ℓt) -∗ ([∗ set] ℓn ∈ s, ℓn ~r~> ℓt) ∗ ℓn ~r~> ℓt.
  Proof.
    iIntros (elem) "map".
    iDestruct (big_sepS_elem_of_acc _ _ ℓn with "map") as "[#reach reins]"; first by set_solver.
    iFrame "#". by iApply "reins".
  Qed.

  Lemma mk_node_info v (ℓn ℓnOut : loc) :
    ℓn ↦□ CONSV v #ℓnOut ==∗ ∃ γn, node_info γn ℓn v ℓnOut ∗ own γn (◯ {[ ℓn ]}).
  Proof.
    iMod (own_alloc ((● {[ ℓn ]}) ⋅ (◯ {[ ℓn ]}))) as (γn) "[auth #frag]"; first by apply auth_both_valid.
    iIntros "#pts". iModIntro. iExists γn. iFrame "#".
    iExistsN. iFrame. rewrite big_sepS_singleton. by iApply reachable_refl.
  Qed.

  Lemma lookup_reach_set ℓn γt ℓt :
    ℓn -r-> γt -∗ tie γt ℓt -∗ ℓn ~r~> ℓt.
  Proof.
    iIntros "#frag". iDestruct 1 as (s) "[auth map]".
    iDestruct (own_valid_2 with "auth frag") as %[subset%gset_included]%auth_both_valid.
    iDestruct (lookup_reach_set_map _ ℓn with "map") as "[??]". set_solver.
    iFrame.
  Qed.

  (* Lemma lookup_reach_set α ℓn ℓt :
    own α (◯ {[ ℓn ]}) -∗ tie α ℓt -∗ tie α ℓt ∗ ℓn ~r~> ℓt.
  Proof.
    iIntros "#frag". iDestruct 1 as (s) "[auth map]".
    iDestruct (own_valid_2 with "auth frag") as %[subset%gset_included]%auth_both_valid.
    iDestruct (lookup_reach_set_map _ ℓn with "map") as "[??]". set_solver.
    iFrame. iExistsN. iFrame.
  Qed. *)

  Lemma insert_reach_set α ℓn ℓt :
    tie α ℓt -∗ ℓn ~r~> ℓt ==∗ tie α ℓt ∗ ℓn -r-> α.
  Proof.
    iDestruct 1 as (s) "[auth map]".
    iIntros "reach".
    iMod (own_update _ _ (● ({[ ℓn ]} ∪ s)) with "auth") as "auth".
    { eapply auth_update_auth. apply gset_local_update. set_solver. }
    iMod (own_update _ _ (● ({[ ℓn ]} ∪ s) ⋅ ◯ {[ ℓn ]}) with "auth") as "[auth frag]".
    { eapply auth_update_core_id. apply _. apply gset_included. set_solver. }
    iFrame. iModIntro.
    destruct (decide (ℓn ∈ s)).
    - iExists _. replace ({[ℓn]} ∪ s) with s; last by set_solver. iFrame.
    - iExists _. iFrame. iApply big_sepS_insert. set_solver. iFrame.
  Qed.

  Lemma node_info_reach γn ℓn v ℓnOut :
    node_info γn ℓn v ℓnOut ==∗
    node_info γn ℓn v ℓnOut ∗ ℓn -r-> γn ∗ isNode ℓn v ℓnOut ∗ ℓn ~r~> ℓn.
  Proof.
    iIntros "[#N r]". iDestruct (reachable_refl with "N") as "nRn".
    iMod (insert_reach_set with "r nRn") as "[r #o]". iModIntro. iFrame "∗#".
  Qed.

  Definition ctx := Datatypes.list (ectxi_language.ectx_item heap_ectxi_lang).

  Lemma load_sentinel (K : ctx) E t (A : lrel Σ) γs γt γl τi (ℓtoS ℓtoT : loc) list ℓlock
   (Hcl : nclose specN ⊆ E) :
   inv queueN (queueInv γs γt γl τi ℓtoS ℓtoT list ℓlock)
   -∗ (∀ ℓs ℓsOut v, isNode ℓs v ℓsOut -∗ ℓs -r-> γs -∗ ℓs -r-> γt
         -∗ ℓs -r-> γl -∗ REL fill K #ℓs << t : A)
   -∗ REL fill K (! #ℓtoS) << t : A.
  Proof.
    iIntros "inv Hlog".
    rel_load_l_atomic.
    iInv queueN as (xs xsₛ) "(isMSQ & Hrest)" "Hclose".
    iDestruct "isMSQ" as (ℓs ? ℓt ? ℓl ?????) "(? & toTPts & ? & infoS & [isNodeT reachT] & [isNodeL reachL] & >#sToT & >#tToL)".
    iModIntro. iExistsN. iFrame. iNext. iIntros "toSPts".
    iMod (node_info_reach with "infoS") as "(infoS & #nToN & isNodeS & sRs)".
    iDestruct (lookup_reach_set with "sToT reachT") as "#sRt".
    iDestruct (lookup_reach_set with "tToL reachL") as "#tRl".
    iDestruct (reachable_trans with "sRt tRl") as "sRl".
    iMod (insert_reach_set with "reachL sRl") as "[reachL #sTol]".
    iMod ("Hclose" with "[-Hlog isNodeS]"). { iNext. iExistsN. iFrame. iExistsN. iFrame "toTPts#∗". }
    iApply ("Hlog" with "[$]"); iFrame "#".
  Qed.

  Lemma load_sentinel_wp E γs γt γl τi (ℓtoS ℓtoT : loc) list ℓlock
   (Hcl : nclose specN ⊆ E) :
   {{{ inv queueN (queueInv γs γt γl τi ℓtoS ℓtoT list ℓlock) }}}
     ! #ℓtoS
   {{{ ℓs v ℓsOut, RET #ℓs; isNode ℓs v ℓsOut ∗ own γs (◯ {[ ℓs ]}) ∗ own γt (◯ {[ ℓs ]}) ∗ own γl (◯ {[ ℓs ]}) }}}.
  Proof.
    iIntros (ϕ) "inv Hϕ".
    iInv queueN as (xs xsₛ) "(isMSQ & Hrest)" "Hclose".
    iDestruct "isMSQ" as (??????????) "(? & toTPts & ? & infoS & [isNodeT reachT] & [isNodeL reachL] & >#sToT & >#tToL)".
    wp_load.
    iMod (node_info_reach with "infoS") as "(infoS & #nToN & isNodeS & sRs)".
    iDestruct (lookup_reach_set with "sToT reachT") as "#sRt".
    iDestruct (lookup_reach_set with "tToL reachL") as "#tRl".
    iDestruct (reachable_trans with "sRt tRl") as "sRl".
    iMod (insert_reach_set with "reachL sRl") as "[reachL #sTol]".
    iMod ("Hclose" with "[-Hϕ isNodeS]"). { iNext. iExistsN. iFrame. iExistsN. iFrame "toTPts#∗". }
    iApply ("Hϕ" with "[$]"); iFrame "#".
  Qed.

  Lemma load_tail (K : ctx) E t (A : lrel Σ) γs γt γl τi (ℓtoS ℓtoT : loc) list ℓlock
    (Hcl : nclose specN ⊆ E) :
    inv queueN (queueInv γs γt γl τi ℓtoS ℓtoT list ℓlock)
    -∗ (∀ (ℓt ℓtOut : loc) v, isNode ℓt v ℓtOut -∗ ℓt -r-> γl -∗ REL fill K #ℓt << t : A)
    -∗ REL fill K (! #ℓtoT) << t : A.
  Proof.
    iIntros "inv Hlog".
    rel_load_l_atomic.
    iInv queueN as (xs xsₛ) "(isMSQ & Hrest)" "Hclose".
    iDestruct "isMSQ" as (ℓs ? ℓt ? ℓl ?????) "(toSPts & ? & ? & infoS & infoT & [isNodeL reachL] & #? & #tToL)".
    iExistsN. iFrame. iModIntro. iNext. iIntros "toTPts".
    iDestruct (lookup_reach_set with "tToL reachL") as "#tRl".
    iDestruct (reachable_is_node with "tRl") as (ℓnOut v) "tPts".
    iMod ("Hclose" with "[-Hlog tPts]"). { iNext. iExistsN. iFrame. iExistsN. iFrame "toSPts". iExistsFrame. }
    iApply ("Hlog" with "[$] [$]").
  Qed.

  (* Load the tail and get reachability information for a node ℓn back. *)
  Lemma load_tail_reach (K : ctx) E t (A : lrel Σ) γs γt γl τi (ℓtoS ℓtoT ℓn : loc) list ℓlock
    (Hcl : nclose specN ⊆ E) :
    inv queueN (queueInv γs γt γl τi ℓtoS ℓtoT list ℓlock)
    -∗ own γt (◯ {[ ℓn ]})
    -∗ (∀ (ℓt ℓtOut : loc) v, isNode ℓt v ℓtOut -∗ ℓn ~r~> ℓt -∗ own γt (◯ {[ ℓt ]})-∗ REL fill K #ℓt << t : A)
    -∗ REL fill K (! #ℓtoT) << t : A.
  Proof.
    iIntros "inv frag Hlog".
    rel_load_l_atomic.
    iInv queueN as (xs xsₛ) "(isMSQ & Hrest)" "Hclose".
    iDestruct "isMSQ" as (ℓs ? ℓt ? ℓl ?????) "(toSPts & ? & nodeList & infoS & tInfo & [isNodeL reachL] & #? & #tToL)".
    iExistsN. iFrame. iModIntro. iNext. iIntros "toTPts".
    iMod (node_info_reach with "tInfo") as "([? reachT] & #tToT & isNodeT & _)".
    iDestruct (lookup_reach_set with "tToL reachL") as "#tRl".
    iDestruct (reachable_is_node with "tRl") as (ℓnOut ?) "N".
    iDestruct (lookup_reach_set with "frag reachT") as "#nRt".
    iMod ("Hclose" with "[-Hlog N nRt]"). { iNext. iExistsN. iFrame. iExistsN. iFrame "nodeList ∗#". }
    iApply ("Hlog" with "[$] [$] [$]").
  Qed.

  Lemma reach_set_advance γn ℓn ℓm :
    tie γn ℓn -∗ ℓn ~r~> ℓm ==∗ tie γn ℓm ∗ ℓm -r-> γn.
  Proof.
    iDestruct 1 as (s) "[auth map]".
    iMod (own_update _ _ (● ({[ ℓm ]} ∪ s)) with "auth") as "auth".
    { eapply auth_update_auth. apply gset_local_update. set_solver. }
    iMod (own_update _ _ (● ({[ ℓm ]} ∪ s) ⋅ ◯ {[ ℓm ]}) with "auth") as "[auth frag]".
    { eapply auth_update_core_id. apply _. apply gset_included. set_solver. }
    iIntros "#rRn".
    iDestruct (reachable_last_is_node with "rRn") as (ℓmOut v) "mPts".
    iModIntro. iFrame. iExistsN. iFrame.
    destruct (decide (ℓm ∈ s)) as [?|elem].
    - replace ({[ℓm]} ∪ s) with s; last by set_solver.
      iApply (big_sepS_impl with "map []"). iModIntro.
      iIntros (x _) "#xRn".
      iApply reachable_trans; done.
    - iApply big_sepS_insert. done.
      iSplitR; first by iApply reachable_refl.
      iApply (big_sepS_impl with "map []"). iModIntro. iIntros (? _) "#xRn".
      iApply reachable_trans; done.
  Qed.

  Lemma load_node (K : ctx) E t (A : lrel Σ) γs γt γl list (ℓtoS ℓnOut ℓn ℓtoT : loc) x ℓlock τi
    (Hcl : nclose specN ⊆ E) :
    inv queueN (queueInv γs γt γl τi ℓtoS ℓtoT list ℓlock)
    -∗ ℓn ↦□ CONSV x #ℓnOut
    -∗ own γs (◯ {[ℓn]})
    -∗ (∀ (v : val), REL fill K v << t : A)
    -∗ REL fill K (! #ℓnOut) << t : A.
  Proof.
    iIntros "#? nPts #nToS post".
    rel_load_persistent_l_atomic'.
    iInv queueN as (xs xsₛ) "(isMSQ & Hrest)" "Hclose".
    rewrite /msQueueInv.
    iDestruct "isMSQ" as (ℓs ℓsOut ????????) "(toSPts & toTPts & (l & >lOutPts & ?) & [>#sPts reachS] & intoT & infoL & >#sToT & >#tToL)".
    iDestruct (lookup_reach_set with "nToS reachS") as "#nRt".
    iDestruct (reachable_unfold with "nRt") as (??) "[>#nPts' [><- | R]]";
      iDestruct (mapsto_agree_persistent_2 with "nPts nPts'") as %[= <- <-].
    - iDestruct (mapsto_agree_persistent_2 with "nPts sPts") as %[= <- <-].
      destruct (xs); simpl.
      * iDestruct "l" as ">->".
        iExistsN. iRight. iExistsN. iFrame. iModIntro. iNext. iIntros "nOutPts".
        iMod ("Hclose" with "[-post]") as "_". { iNext. iExistsFrame. }
        done.
      * iDestruct "l" as (??) "(>#nOutPts & ? & ?)".
        iExistsN. iLeft. iFrame "#". iModIntro. iNext.
        iMod ("Hclose" with "[-post]"). { iNext. iExistsFrame. }
        done.
    - iDestruct "R" as (?) "[>#nOutPts ?]".
      iExistsN. iLeft. iFrame "#". iModIntro. iNext.
      iMod ("Hclose" with "[-post]").
      { iNext. iExistsN. iFrame. iExistsN. iFrame "toSPts toTPts infoL ∗#". }
      done.
  Qed.

  (* Suppose ℓt can reach ℓl and that ℓl has no successors. Then any other node
     ℓn which is reachable from ℓt can reach ℓl.
     ℓt ~r~> ℓn ~r~> ℓl
   *)
  Lemma reach_end_node (ℓt ℓl ℓlOut ℓn ℓn' : loc) v :
    ℓt ~r~> ℓl -∗ ℓt ~r~> ℓn -∗ ℓl ↦□ SOMEV (v, #ℓlOut) -∗ ℓlOut ↦ #ℓn' -∗ ℓn' ↦□ NONEV
    -∗ ℓlOut ↦ #ℓn' ∗ ℓn ~r~> ℓl.
  Proof.
    iIntros "tRl tRn lPts lOutPts nPts".
    iDestruct "tRl" as (n) "R".
    iInduction n as [|n'] "IH" forall (ℓt).
    - iDestruct "R" as (?v ℓlOut') "(lPts' & ->)".
      iDestruct (reachable_nil_eq with "tRn lPts lOutPts nPts") as %->.
      iFrame.
    - iDestruct "R" as (?v ℓtOut) "(tPts & next)". fold reachable_l.
      iDestruct "next" as (?) "[tOutPts R]".
      iDestruct (reachable_unfold with "tRn") as (?v ℓtOut') "(tPts' & [%eq|next])".
      * subst.
        iDestruct (mapsto_agree_persistent_2 with "tPts tPts'") as %[= <- <-].
        iFrame. iExists (S n'). iExistsFrame.
      * iDestruct "next" as (ℓp') "[tOutPts' pRn]".
        iDestruct (mapsto_agree_persistent_2 with "tPts tPts'") as %[= <- <-].
        iDestruct (mapsto_agree_persistent_2 with "tOutPts tOutPts'") as %[= ->].
        iDestruct ("IH" with "R pRn lPts lOutPts nPts") as "[lOutPts nRl]".
        iFrame.
  Qed.

  Lemma advance_tail_pointer (K : ctx) γs γt γl τi ℓtoS ℓtoT list l ℓt ℓn t A :
    inv queueN (queueInv γs γt γl τi ℓtoS ℓtoT list l)
    -∗ ℓt ~r~> ℓn
    -∗ (∀ (v : val), REL fill K v << t : A)
    -∗ REL fill K (Snd (CmpXchg #ℓtoT #ℓt #ℓn)) << t : A.
  Proof.
    iIntros "#inv #tRn post".
    iDestruct (reachable_last_is_node with "tRn") as (??) "nIsNode".
    rel_cmpxchg_l_atomic.
    iInv queueN as (xs xsₛ) "(isMSQ & Hsq & Hlink)" "Hclose".
    iDestruct "isMSQ" as (ℓs ℓsOut ????????)
      "(toSPts & toTPts & (nodeList & lOutPts & #n'Pts) & sInfo & [? reachT] & [#lIsNode reachL] & >#sToT & >#tToL)".
    iExistsN. iFrame "toTPts". iModIntro.
    iSplit.
    -- iIntros "_". iNext. iIntros "toTPts". rel_pures_l.
      iMod ("Hclose" with "[-post]") as "_".
      { iNext. iExistsN. iFrame. iExistsN. iFrame "toSPts". iFrame. iFrame "#". }
      iApply "post".
    -- iIntros ([= ->]). iNext. iIntros "toTPts". rel_pures_l.
      iDestruct (lookup_reach_set with "tToL reachL") as "#tRl".
      iDestruct (reach_end_node with "tRl tRn lIsNode lOutPts n'Pts") as "[lOutPts nRl]".
      iMod (insert_reach_set with "reachL nRl") as "[reachL #nTol]".
      iMod (reach_set_advance with "reachT tRn") as "[reachT #nR]".
      iMod ("Hclose" with "[-post]") as "_".
      { iNext. iExistsN. iFrame. iExistsN. iFrame "toSPts". iFrame "∗#". }
      iApply "post".
  Qed.

  Lemma changed_last_pointer xs x (ℓlOut ℓsOut ℓnil ℓtoTail ℓnode ℓn : loc) :
    ℓnil ↦□ NILV -∗
    ℓtoTail ↦ #ℓnil -∗
    ℓnode ↦ CONSV (InjRV x) #ℓtoTail -∗
    isMSQueue ℓlOut ℓsOut xs -∗
    ℓn ↦□ NONEV -∗
    ℓlOut ↦ #ℓnode ==∗
      ℓnode ↦□ CONSV (InjRV x) #ℓtoTail ∗
      isMSQueue ℓtoTail ℓsOut (xs ++ [x]) ∗
      ℓtoTail ↦ #ℓnil ∗
      ℓlOut ↦□ #ℓnode.
  Proof.
    iIntros "#nilPts toTailPts nodePts nodeList #nPts lOutPts".
    iInduction xs as [|x' xs'] "IH" forall (ℓsOut).
    - iDestruct "nodeList" as %<-.
      iFrame.
      iMod (mapsto_persist with "nodePts") as "#nodePts".
      iMod (mapsto_persist with "lOutPts") as "#lOutPts".
      iModIntro. iFrame "#". iExists _, _. iFrame "#". auto.
    - iDestruct "nodeList" as (??) "(? & ? & nodeList)".
      iMod ("IH" with "toTailPts nodePts nodeList lOutPts") as "(?&?&?)".
      iModIntro. iFrame. iExistsN. iFrame.
  Qed.

  (* If ℓn can reach ℓl, and ℓlOut points to ℓo then there exists an ℓm which ℓnOut points to.
     It is either ℓo or some other node. *)
  Lemma reach_node (ℓn ℓnOut ℓl ℓlOut ℓo : loc) v1 v2 :
    ℓn ~r~> ℓl
    -∗ ℓn ↦□ CONSV v1 #ℓnOut
    -∗ ℓl ↦□ CONSV v2 #ℓlOut
    -∗ ℓlOut ↦ #ℓo
    -∗ ∃ (ℓm : loc),
        (∃ v (ℓmOut : loc), ℓnOut ↦□ #ℓm ∗ ℓm ↦□ CONSV v #ℓmOut ∗ ℓlOut ↦ #ℓo) ∨
        (⌜ℓn = ℓl⌝ ∗ ⌜ℓnOut = ℓlOut⌝ ∗ ⌜ℓm = ℓo⌝ ∗ ℓnOut ↦ #ℓo).
  Proof.
    iIntros "nRl nPts lPts lOutPts".
    iDestruct (reachable_unfold with "nRl") as (??) "[nPts' [-> | R]]";
    iDestruct (mapsto_agree_persistent_2 with "nPts nPts'") as %[= <- <-].
    - iDestruct (mapsto_agree_persistent_2 with "nPts lPts") as %[= <- <-].
      iExists ℓo. iRight. iFrame. done.
    - iDestruct "R" as (ℓm) "[nOutPts mRl]".
      iDestruct (reachable_is_node with "mRl") as (?v ?) "mPts".
      iExists _. iLeft. iExists _, _. iFrame.
  Qed.

  Lemma queue_refinement :
    ⊢ REL MS_queue_no_snapshot << CG_queue : ∀ A, (() → () + A) * (A → ()).
  Proof.
    iApply refines_forall. iModIntro. iIntros (τi).
    (* Step through the initialization of the implementation. *)
    rel_alloc_l nil as "nilPts". iMod (mapsto_persist with "nilPts") as "#nilPts".
    rel_alloc_l tail as "tailPts".
    rel_alloc_l ℓs as "sPts". iMod (mapsto_persist with "sPts") as "#sPts".
    rel_alloc_l ℓtoT as "toTPts".
    rel_alloc_l ℓtoS as "toSPts".
    rel_pures_l.
    (* Step through the initialization of the specification. *)
    rel_alloc_r list as "listPts".
    rel_pures_r.
    rel_apply_r refines_newlock_r. iIntros (l) "lofal".
    rel_pure_r.
    rel_pure_r.
    rel_pure_r.
    rel_pure_r.

    (* We now want to allocate the invariant for the queue. To do that we first
       establish the node invariant for the empty node we have created. *)
    iMod (mk_node_info with "sPts") as (γs) "[infoS _]".
    iMod (mk_node_info with "sPts") as (γt) "[infoT sToT]".
    iMod (mk_node_info with "sPts") as (γl) "[infoL sToL]".
    iMod (inv_alloc queueN _ (queueInv γs γt γl τi ℓtoS ℓtoT list _) with "[-]")
      as "#Hinv".
    { iNext. iExists [], []. simpl. iFrame. iExistsN. iFrame "∗#". done. }
    iClear "nilPts sPts". clear ℓs tail nil.
    iApply refines_pair.
    - (* dequeue *)
      rewrite /MS_dequeue_no_snapshot /CG_dequeue.
      rel_pure_l. rel_pure_l. rel_pure_l. rel_pure_l.
      iApply refines_arrow_val. iModIntro.
      iLöb as "IH".
      (* Load the sentinel *)
      iIntros (? ? [-> ->]). rel_pures_l. clear.
      rel_apply_l (load_sentinel with "[$]"); first done.
      iIntros (ℓs ℓsOut ?v) "#sPts #sRs #sToT #sToL".
      rel_pures_l.
      (* Load the tail *)
      rel_apply_l (load_tail_reach with "[$] sToT"); first done.
      iIntros (ℓt ℓtOut ?) "_ sRt #tReach".
      rel_pures_l.
      rel_load_l.
      rel_rec_l. rel_pures_l.
      (* The points-to predicate for ℓsOut is inside the invariant. *)
      rel_load_persistent_l_atomic'.
      iInv queueN as (xs xsₛ) "(isMSQ & Hsq & Hlink)" "Hclose".
      iDestruct "isMSQ" as (ℓs' ℓs'Out ? ? ℓl' ℓlOut ????)
        "(? & toTPts & nodeList & [s'IsNode reachS] & [isNodeT reachT] & [>lPts reachL] & >#s'ToT & >#t'ToL)".
      (* The old sentinel can reach the current sentinel, tail, and last node. *)
      iDestruct (lookup_reach_set with "sRs reachS") as " #sRs'".
      iDestruct (lookup_reach_set with "sToL reachL") as "#sRl".
      iDestruct (lookup_reach_set with "sToT reachT") as "#sRt'".
      (* We know that the current last node, ℓl', is reachable from, ℓs, the
         sentinel we read earlier. This means that either ℓs is equal to ℓl or
         ℓs has a succssor node. *)
      iDestruct (reachable_unfold with "sRl") as (?v ℓsOut') "[>sPts'' [>%eq|next]]".
      + (* In this case ℓs is equal to ℓl'. This implies that the queue is empty,
           i.e., that xs is the empty list *)
        subst.
        iDestruct (mapsto_agree_persistent_2 with "sPts lPts") as %[= -> ->].
        iClear "sPts".
        iDestruct "nodeList" as "(nodeList & >lOutPts & #nPts)".
        iExists _. iRight. iExists _. iFrame "lOutPts". iModIntro. iNext. iIntros "lOutPts".
      
        (* We conclude that ℓl' = ℓs' = ℓt*)
        iDestruct (reachable_nil_eq with "sRs' lPts lOutPts nPts") as %lEqs'.
        iDestruct (reachable_nil_eq with "sRt lPts lOutPts nPts") as %lEqt.
        rewrite -lEqs'.
        rewrite -lEqt.
      
        iDestruct (mapsto_agree_persistent_2 with "lPts s'IsNode") as %[= <- <-].
        iDestruct (is_ms_queue_eq with "nodeList lOutPts nPts") as %->.

        destruct xsₛ as [|x' xsₛ'].
        2: { iDestruct (big_sepL2_length with "Hlink") as %[=]. }
        simpl.
        rel_rec_r.
        rel_apply_r (refines_CG_dequeue_nil_r with "Hsq").
        iIntros "Hsq".
        iMod ("Hclose" with "[-nPts]") as "_".
        { iNext. iExists [], []. simpl. iFrame. iExistsN. iFrame. iFrame "#". }
        rel_pures_l.
        rewrite bool_decide_eq_true_2 //. rel_pures_l.
        subst.
        rel_load_l.
        rel_pures_l.
        rel_values. iModIntro. iExists _, _. auto.
      + (* The sentinel has a successor. *)
        iDestruct (mapsto_agree_persistent_2 with "sPts sPts''") as %[= <- <-].
        iDestruct "next" as (ℓn) "[>#sOutPts nRl]".
        iExistsN. iLeft. iFrame "sOutPts". iModIntro. iNext.
        iDestruct (reachable_unfold with "nRl") as (?v ?) "[#nPts _]".
        iMod ("Hclose" with "[-nPts sPts sOutPts sRt]") as "_".
        { iNext. iExistsN. iFrame. iExistsN. iFrame "toTPts nodeList #∗". }
        rel_pures_l.
        (* The if-statement on line D8. *)
        destruct (decide (ℓs = ℓt)) as [->|neq].
        { (* In this case the tail pointer is outdated. *)
          rewrite bool_decide_eq_true_2 //. rel_pures_l.
          rel_load_l. rel_pures_l.
          iDestruct (reachable_succ with "sPts sOutPts nPts") as "tRn".
          rel_apply_l (advance_tail_pointer with "Hinv tRn"). iIntros (?).
          rel_pure_l. rel_pure_l. by iApply "IH". }
        rewrite bool_decide_eq_false_2. 2: { intros [=]. done. }
        rel_pures_l.
        rel_cmpxchg_l_atomic.
        iInv queueN as (xs2 xsₛ2) "(isMSQ & Hsq & Hlink)" "Hclose".
        iClear "s'ToT t'ToL sRs' sRl nRl". clear ℓs' ℓs'Out ℓl' ℓlOut.
        rewrite /msQueueInv.
        iDestruct "isMSQ" as (ℓs2 ℓs2Out ℓt2 ? ℓl ℓlOut ????)
          "(toSPts & toTPts & (nodeList & lOutPts & nilPts) & [sPts' reachS] & [isNodeT reachT] & [>lIsNode reachL] & >#s2ToT & >#t'ToL)".
        iExists _. iFrame "toSPts".
        iModIntro. iSplit.
        2: {
          (* The queue still points to the same sentinel that we read earlier--the CAS succeeds *)
          iIntros ([= ->]). iNext. iIntros "toSPts".
          iDestruct (mapsto_agree_persistent_2 with "sPts sPts'") as %[= <- <-].
          (* xs2 is not necessarily equal to xs, but, since the CAS succeeded,
             it still has xs as a prefix. And since we know that xs is a cons xs2
             must also be a cons with the same element. *)
          destruct xs2 as [|x2' xs2']; simpl.
          { iDestruct "nodeList" as %->.
            iDestruct (mapsto_agree_persistent with "sOutPts lOutPts") as %[= ->].
            iDestruct (mapsto_agree_persistent_2 with "nilPts nPts") as %[=]. }
          destruct xsₛ2 as [|xₛ2' xsₛ2']; first done.
          (* Since we now know that xs2 is a cons we can get the first node out
             of the representation predicate. *)
          iDestruct "nodeList" as (ℓfstN ℓnext) "(sOutPts' & tailPts & nodeList)".
          iDestruct (mapsto_agree_persistent_2 with "sOutPts sOutPts'") as %[= <-].
          iDestruct (mapsto_agree_persistent_2 with "nPts tailPts") as %[= -> ->].
          iDestruct "Hlink" as "[Hτi Hlink]".
          rel_pures_l.
          (* We step through the specificaion code. *)
          rel_rec_r.
          rel_apply_r (refines_CG_dequeue_cons_r with "Hsq"). iIntros "Hsq".
          (* We need to ensure that the new sentinel can still reach the tail.
          This is slightly tricky. *)
          iDestruct (lookup_reach_set with "tReach reachT") as "#tRt2".
          iDestruct (reachable_transp with "sRt sPts' sOutPts nPts") as "nRt"; first assumption.
          iDestruct (reachable_trans with "nRt tRt2") as "nRt2".
          iMod (insert_reach_set with "reachT nRt2") as "[reachT #nToT]".
          (* We must "move the reach set forward" ℓs forward and get: tie γs ℓfstN *)
          iDestruct (reachable_succ with "sPts sOutPts nPts") as "sRn".
          iMod (reach_set_advance with "reachS sRn") as "[reachS #nToS]".
          (* We are now ready to reestablish the invariant. *)
          iMod ("Hclose" with "[-Hτi]") as "_".
          { iNext. iExistsN. iFrame. iExistsN. iFrame "toSPts ∗#". }
          rel_load_l. rel_rec_l. rel_pures_l. rel_rec_l. rel_pures_l.
          rel_values. iExistsN. auto. }
        * (* CAS failed, the queue no longer points to the same sentinel. *)
          iIntros (Hneq). iNext. iIntros "toSPts".
          iMod ("Hclose" with "[-IH]") as "_".
          { iNext. iExistsN. iFrame. iExistsN. iFrame "toTPts ∗#". }
          rel_pure_l. rel_pure_l.
          by iApply "IH".
    - (* enqueue *)
      rewrite /MS_enqueue_no_snapshot /CG_enqueue.
      rel_pures_l.
      rel_pures_r.
      iApply refines_arrow_val. iModIntro.
      iIntros (x x') "#Hrel". rel_pures_l.
      (* We allocate the new node *)
      rel_alloc_l ℓnil as "nilPts". iMod (mapsto_persist with "nilPts") as "#nilPts".
      rel_alloc_l ℓtoTail as "tailPts".
      rel_alloc_l ℓnode as "nodePts".
      rel_pure_l. rel_pure_l. rel_pure_l.
      iLöb as "IH".
      rel_pures_l.
      rel_apply_l (load_tail with "[$]"); first done.
      iIntros (ℓt ℓtOutPts ?v) "tPts #tToL".
      rel_pures_l.
      (* Lookup node *)
      rel_load_l.
      rel_rec_l.
      rel_pures_l.
      (* Lookup node *)
      rel_load_persistent_l_atomic'.
      iInv queueN as (xs xsₛ) "(isMSQ & Hsq & Hlink)" "Hclose".
      iDestruct "isMSQ" as (ℓs ℓsOut ℓt' ? ℓl ℓlOut ????) "(toSPts & toTPts & (nodeList & lOutPts & #nilPts') & [sIsNode reachS] & tInfo & [>lPts reachL] & >#sToT & >#t'ToL)".
      iDestruct (lookup_reach_set with "tToL reachL") as "#tRl".
      iDestruct (reachable_unfold with "tRl") as (??) "[>tPts' disj]".
      iDestruct (mapsto_agree_persistent_2 with "tPts tPts'") as %[= <- <-].
      iClear "tPts'".
      iDestruct "disj" as "[>-> | right]".
      + (* We are at the end and can attempt inserting our node. *)
        iDestruct (mapsto_agree_persistent_2 with "lPts tPts") as %[= <- <-].
        iExistsN. iRight. iExistsN. iFrame "lOutPts". iModIntro. iNext. iIntros "lOutPts".
        iMod ("Hclose" with "[-tailPts nodePts tPts]") as "_".
        { iNext. iExistsN. iFrame. iExistsN. iFrame "toSPts toTPts lOutPts". iFrame "#∗". }
        clear xs xsₛ.
        rel_pures_l.
        (* rewrite bool_decide_eq_true_2 //. *)
        rel_pures_l.
        rel_load_l.
        rel_pures_l.
        rel_cmpxchg_persistent_l_atomic.
        (* We must open the invariant, case on whether ℓ is equal to ℓ2, and
           extract that ℓ is the last node. *)
        iInv queueN as (xs xsₛ) "(isMSQ & Hsq & Hlink)" "Hclose".
        iDestruct "isMSQ" as (ℓs2 ℓs2Out ℓt2 ? ℓl' ℓl'Out ????)
          "(toSPts & toTPts & (#nodeList & >l'OutPts & #nilPts2) & [s2IsNode reachS] & tInfo & [>#l'Pts reachL] & >#s2ToT & >#t2ToL)".
        iDestruct (lookup_reach_set with "tToL reachL") as "#tRl'".
        iDestruct (reach_node with "tRl' [>tPts//] [>l'Pts//] [>l'OutPts//]") as (ℓm) "disj".
        iExists #ℓm. iModIntro. iSplit.
        2: { (* We are still at the end. *)
          iIntros ([= Eq]). subst.
          iDestruct "disj" as "[R|R]".
          { iDestruct "R" as (??) "(_ & >nilPts'' & _)".
            iDestruct (mapsto_agree_persistent_2 with "nilPts' nilPts''") as %[=]. }
          iDestruct "R" as "(><- & ><- & ><- & >l'OutPts)".
          iNext.
          iFrame "l'OutPts". iIntros "lOutPts".
          iRename "nodeList" into "oldNodeList".
          iMod (changed_last_pointer with "nilPts tailPts nodePts oldNodeList nilPts lOutPts")
            as "(#nodePts & nodeList & toTailPts & lOutPts)".
          (* We have successfully added a successor to ℓt and can conclude that ℓnode is reachable from ℓt. *)
          iDestruct (reachable_succ with "l'Pts lOutPts nodePts") as "#tRnode".
          iMod (reach_set_advance with "reachL tRnode") as "[reachL _]".
          fold CG_enqueue.
          rel_rec_r.
          rel_apply_r (refines_CG_enqueue_r with "Hsq").
          iIntros "Hsq".
          iMod ("Hclose" with "[-Hinv]") as "_".
          { iNext. iExists (xs ++ [x]), (xsₛ ++ [x']).
            iFrame. iSplitL. iExistsN. iFrame "toSPts toTPts nodeList".
            iFrame "toTailPts". iFrame "#∗". by iApply big_sepL2_singleton. }
          clear.
          rel_pures_l.
          (* We now attempt to update the tail pointer *)
          rel_apply_l (advance_tail_pointer with "Hinv [$]"). iIntros (?).
          rel_pures_l. rel_values. }
        * (* Another thread changed the end in the meantime. *)
          iIntros "Neq".
          iDestruct "disj" as "[R|R]".
          --
            iDestruct "R" as (??) "(lOutPts & mPts & l'OutPts)".
            iLeft. iFrame "lOutPts". iNext.
            iMod ("Hclose" with "[-IH tailPts nodePts]") as "_".
            { iNext. iExistsN. iFrame. iExistsN. iFrame "toTPts toSPts ∗#". }
            rel_pure_l. rel_pure_l.
            iApply ("IH" with "tailPts nodePts").
          --
            iRight. iExists _.
            iDestruct "R" as "(>-> & >-> & >-> & lOutPts)".
            iFrame "lOutPts". iNext. iIntros "lOutPts".
            iMod ("Hclose" with "[-IH nilPts tailPts nodePts]") as "_".
            { iNext. iExistsFrame. }
            rel_pure_l. rel_pure_l.
            iApply ("IH" with "tailPts nodePts").
      + (* We are not at the end yet, keep on going. *)
        iDestruct "right" as (ℓn) "[>#tOutPts nRl]". (* FIXME: weird name *)
        iExistsN. iLeft. iFrame "tOutPts". iModIntro. iNext.
        iDestruct (reachable_is_node with "nRl") as (ℓnOut ?v) "#nPts".
        iDestruct (reachable_succ with "tPts tOutPts nPts") as "#tRn".
        rel_pures_l.
        iMod ("Hclose" with "[-nilPts tailPts nodePts]") as "_".
        { iNext. iExistsN. iFrame. iExistsN. iFrame "toSPts toTPts nodeList ∗#". }
        rel_pures_l.
        rel_load_l. rel_pures_l.
        rel_apply_l (advance_tail_pointer with "Hinv tRn"). iIntros (?).
        rel_pure_l.
        rel_pure_l.
        iApply ("IH" with "tailPts nodePts").
  Qed.

End queue_refinement.

Open Scope nat.
Theorem MS_no_snapshot_ctx_refines_CG :
  ∅ ⊨ MS_queue_no_snapshot ≤ctx≤ CG_queue :
    ∀: (TUnit → TUnit + TVar 0) * (TVar 0 → TUnit).
Proof.
  pose (Σ := #[relocΣ; queueΣ]).
  eapply (refines_sound Σ).
  iIntros (? Δ).
  iApply queue_refinement.
Qed.
