From reloc Require Import reloc lib.lock.
From msqueue Require Import common extra_rules CG_queue MS_queue MS_queue_no_snapshot.
From msqueue Require Import CG_refines_MS_no_snapshot MS_no_snapshot_refines_CG
                            MS_refines_CG CG_refines_MS.

Section equivalence.

  (* The MS-queue is contextually equivalent to the coarse-grained queue. *)
  Lemma MS_ctx_equivalent_FG :
    ∅ ⊨ MS_queue_original =ctx= CG_queue : ∀: (TUnit → TUnit + TVar 0) * (TVar 0 → TUnit).
  Proof.
    split.
    - apply MS_ctx_refines_CG.
    - apply CG_ctx_refines_MS.
  Qed.

  (* The MS-queue without consistent snapshots is contextually equivalent to the coarse-grained queue. *)
  Lemma MS_no_snapshot_ctx_equivalent_FG :
    ∅ ⊨ MS_queue_no_snapshot =ctx= CG_queue : ∀: (TUnit → TUnit + TVar 0) * (TVar 0 → TUnit).
  Proof.
    split.
    - apply MS_no_snapshot_ctx_refines_CG.
    - apply CG_ctx_refines_MS_no_snapshot.
  Qed.

  (* The MS-queue with consistent snapshots is contextually equivalent to one without. *)
  Lemma MS_original_ctx_equivalent_MS_no_snapshot :
    ∅ ⊨ MS_queue_original =ctx= MS_queue_no_snapshot : ∀: (TUnit → TUnit + TVar 0) * (TVar 0 → TUnit).
  Proof.
    split.
    - eapply ctx_refines_transitive.
      apply MS_ctx_refines_CG.
      apply CG_ctx_refines_MS_no_snapshot.
    - eapply ctx_refines_transitive.
      apply MS_no_snapshot_ctx_refines_CG.
      apply CG_ctx_refines_MS.
  Qed.

End equivalence.