(* This file contains extra rulse that we add to Iris. *)

From reloc Require Import reloc lib.lock.
From msqueue Require Import common CG_queue MS_queue.
From iris.algebra Require Import numbers csum excl auth list gmap gset.
From iris_string_ident Require Import ltac2_string_ident.
From iris.bi.lib Require Export fixpoint.
From iris_string_ident Require Import ltac2_string_ident.

Section extra_rules.

  Context `{relocG Σ}.

  Lemma refines_load_alt K E l t A :
    (|={⊤,E}=> ∃ v' q,
      ▷(l ↦{q} v') ∗
      ▷(l ↦{q} v' -∗ (REL fill K (of_val v') << t @ E : A)))%I
    -∗ REL fill K (! #l) << t : A.
  Proof.
    iIntros "Hlog".
    iApply refines_atomic_l; auto.
    iMod "Hlog" as (v' q) "[Hl Hlog]". iModIntro.
    iApply (wp_load with "Hl"); auto.
  Qed.

  (* A load rule that supports the persistent points-to predicate. *)
  Lemma refines_load_persistent_l' K E l t A :
    (|={⊤,E}=> ∃ v,
      (l ↦□ v ∗ ▷(REL fill K (of_val v) << t @ E : A)
      ∨ (∃ q, ▷(l ↦{q} v) ∗
              ▷(l ↦{q} v -∗ (REL fill K (of_val v) << t @ E : A)))))%I
    -∗ REL fill K (! #l) << t : A.
  Proof.
    iIntros "Hlog".
    iApply refines_atomic_l; auto.
    iMod "Hlog" as (v) "[[#Hl Hlog] | Hlog]".
    - iModIntro. iApply (wp_load_persistent with "Hl"); auto.
    - iDestruct "Hlog" as (q) "[Hl Hlog]".
      iModIntro. iApply (wp_load with "Hl"); auto.
  Qed.

  Lemma refines_cmpxchg_persistent_l K E l e1 e2 v1 v2 t A :
    IntoVal e1 v1 →
    IntoVal e2 v2 →
    val_is_unboxed v1 →
    (|={⊤,E}=> ∃ v,
     (⌜v ≠ v1⌝ -∗ ▷ (l ↦□ v ∗ REL fill K (of_val (v, #false)) << t @ E : A) ∨
                    ∃ q, ▷ (l ↦{q} v ∗ (l ↦{q} v -∗ REL fill K (of_val (v, #false)) << t @ E : A))) ∧ 
     (⌜v = v1⌝ -∗ ▷ (l ↦ v ∗ (l ↦ v2 -∗ REL fill K (of_val (v, #true)) << t @ E : A))))
    -∗ REL fill K (CmpXchg #l e1 e2) << t : A.
  Proof.
    iIntros (Heq Heq' ?) "Hlog". rewrite -Heq -Heq'.
    iApply (refines_atomic_l E).
    iMod "Hlog" as (v) "Hlog".
    destruct (decide (v = v1)).
    - (* CmpXchg successful *) subst.
      iDestruct "Hlog" as "[_ Hlog]".
      iDestruct ("Hlog" with "[% //]") as "[Hl Hlog]".
      iApply (wp_cmpxchg_suc with "Hl"); eauto.
      by right.
    - (* CmpXchg failed *)
      iDestruct "Hlog" as "[Hlog _]".
      iDestruct ("Hlog" with "[% //]") as "[[Hl Hlog]|Hlog]".
      * iApply (wp_cmpxchg_persistent_fail with "Hl"); eauto. by right.
      * iDestruct "Hlog" as (q) "[Hl Hlog]".
        iApply (wp_cmpxchg_fail with "Hl"); eauto. by right.
  Qed.

  (* Rule for atomic resolution of a prophecy variables. This is a simplified
  version of the rule `refines_resolveatomic_l`. This version is sufficient for
  our purpose and matches the rule presented in the paper. *)
  Lemma refines_resolveatomic_simple_l K (p : proph_id) e w vs t A :
    Atomic StronglyAtomic e →
    to_val e = None →
    (proph p vs ∗
            WP e {{ v, ∀ (vs' : list (val*val)), ⌜vs = (v,w)::vs'⌝ -∗
              proph p vs' -∗
              REL fill K (of_val v) << t : A }})%I
    -∗ REL fill K (Resolve e #p w) << t : A.
  Proof.
    iIntros (??) "Hlog".
    iApply refines_atomic_l; auto.
    { apply strongly_atomic_atomic. apply _. }
    iDestruct "Hlog" as "[Hp Hlog]". iModIntro.
    iApply (wp_resolve with "Hp"); auto.
  Qed.

  Local Hint Resolve tpool_lookup : core.
  Local Hint Resolve tpool_lookup_Some : core.
  Local Hint Resolve to_tpool_insert : core.
  Local Hint Resolve to_tpool_insert' : core.
  Local Hint Resolve tpool_singleton_included : core.
  
  Lemma step_resolve_load E j K l q v (p : proph_id) w :
    nclose specN ⊆ E →
    spec_ctx ∗ l ↦ₛ{q} v ∗ j ⤇ fill K (Resolve (Load #l) #p (of_val w)) ={E}=∗
    j ⤇ fill K v ∗ l ↦ₛ{q} v.
  Proof.
    iIntros (?) "(#Hinv & Hl & Hj)".
    rewrite /spec_ctx tpool_mapsto_eq /tpool_mapsto_def /=.
    iDestruct "Hinv" as (ρ) "Hinv".
    iInv specN as (tp σ) ">[Hown %]" "Hclose".
    iDestruct (own_valid_2 with "Hown Hj")
      as %[[?%tpool_singleton_included' _]%prod_included _]%auth_both_valid.
    iMod (own_update_2 with "Hown Hj") as "[Hown Hj]".
    { by eapply auth_update, prod_local_update_1,
         singleton_local_update, (exclusive_local_update _ (Excl (fill K v))). }
    (* Get the points-to predicate. *)
    rewrite heapS_mapsto_eq /heapS_mapsto_def /=.
    iDestruct (own_valid_2 with "Hown Hl")
      as %[[? ?%gen_heap_singleton_included]%prod_included ?]%auth_both_valid.
    iFrame. iApply "Hclose". iNext.
    iExists (<[j:=fill K v]> tp), σ.
    rewrite to_tpool_insert'; last eauto. iFrame. iPureIntro.
    eapply rtc_r, step_insert_no_fork; eauto.
    econstructor. econstructor. auto.
  Qed.

  Lemma refines_resolve_load_r E K l v t (p : proph_id) w A
    (Hmasked : nclose specN ⊆ E) :
    l ↦ₛ v -∗ 
    (l ↦ₛ v -∗ REL t << fill K (of_val v) @ E : A)%I
    -∗ REL t << fill K (Resolve (! #l) #p (of_val w)) @ E : A.
  Proof.
    iIntros "Hl Hlog".
    pose (Φ := (fun w => ⌜w = v⌝ ∗ l ↦ₛ v)%I).
    iApply (refines_step_r Φ with "[Hl]"); simpl; auto.
    { cbv[Φ].
      iIntros (j K') "#Hs Hj /=".
      iMod (step_resolve_load with "[$Hs $Hl $Hj]") as "[Hj Hl]". done.
      iModIntro. iExists _. iFrame. eauto. }
    iIntros (v') "He'". iDestruct "He'" as "[-> Hl]".
    by iApply "Hlog".
  Qed.

End extra_rules.

Tactic Notation "rel_load_l_atomic" := rel_apply_l refines_load_alt.

Tactic Notation "rel_load_persistent_l_atomic'" := rel_apply_l refines_load_persistent_l'.

Tactic Notation "rel_cmpxchg_persistent_l_atomic" := rel_apply_l refines_cmpxchg_persistent_l.
