Definition MS_dequeue_original : val :=
  λ: "toHead" "toTail", rec: "loop" <> :=
    let: "head" := !"toHead" in                     (* Load the head/sentinel *)
    let: "tail" := !"toTail" in                     (* Load the tail *)
    let: "toNext" := Snd (getValue !"head") in      (* Get the pair in the head/sentinel *)
    let: "next" := !"toNext" in                     (* Read the pointer out of the head . *)
    if: "head" = !"toHead"                          (* Is `head` still up-to-date? *)
    then
      if: "head" = "tail"                           (* If this is false the queue is non-empty *)
      then
        match: rec_unfold !"next" with              (* Does the head point to nil/null? *)
          NONE => NONE                              (* The queue is empty *)
        | SOME <> => CAS "toTail" "tail" "next"     (* Tail is outdated, try to advance it *)
        end
      else
        if: CAS "toHead" "head" "next"              (* Attempt to advance the head pointer *)
        then SOME (getValue (Fst (getValue !"next"))) (* Since we're in this case, next is not nil *)
        else "loop" #()
    else "loop" #()                                 (* `head` is outdated, try again *).

Definition MS_enqueue_original : val :=
  λ: "toTail" "x",
    let: "node" := ref (SOME (SOME "x", ref (ref NONE))) in (* Allocate new node *)
    (rec: "loop" <> :=
      let: "tail" := !"toTail" in                   (* Load the tail pointer *)
      let: "toNext" := Snd (getValue !"tail") in    (* Get pointer from tail into next node *)
      let: "next" := !"toNext" in
      if: "tail" = !"toTail"                        (* Is `tail` still up-to-date? *)
      then
        match: !"next" with
          NONE =>                                   (* The tail is nil, we can try to insert *)
            if: CAS "toNext" "next" "node"
            then CAS "toTail" "tail" "node"         (* Enqueue done, try to move the tail-pointer forward *)
            else "loop" #()                         (* Insertion failed, we try again *)
        | SOME <> =>                                (* `toTail` didn't point to the last node *)
          CAS "toTail" "tail" "next" ;; "loop" #()  (* Try to move the tail-pointer forward, then retry *)
        end
      else "loop" #()
    ) #().

Definition MS_queue_original : val :=
  Λ: let: "node" := ref (SOME (NONE, ref (ref NONE))) in
    let: "toTail" := ref "node" in
    let: "toHead" := ref "node" in
    (MS_dequeue_original "toHead" "toTail", MS_enqueue_original "toTail").